#ifndef SNEeSe_CPU_i
#define SNEeSe_CPU_i

#include "regs.i"

#define _P_B        _65c816_P_B // Flags Register - See information in CPU.S!
#define _P_W        _65c816_P_W
#define _Base       _65c816_Base    // Base pointer to register set
#define _Cycles     _65c816_Cycles  // Cycle counter
#define _NativePC   _65c816_NativePC
#define _P  __65816_P
#define _A  __65816_A
#define _X  __65816_X
#define _Y  __65816_Y
#define _S  __65816_S
#define _D  __65816_D
#define _PC __65816_PC
#define _DB_Shifted __65816_DB_Shifted
#define _PB_Shifted __65816_PB_Shifted
#define _DB (_DB_Shifted+2)
#define _PB (_PB_Shifted+2)

#define EventTrip _EventTrip
#define LineCycles _LineCycles

//  True 65816 layout, Native Mode    = |E|N|V|M|X|D|I|Z|C|
//  True 65816 layout, Emulation Mode = |E|N|V|1|B|D|I|Z|C|

// These are the bits for flag set/clr operations
// |E|N|Z|I|D|X|M|V|C|
//           |B|1|

SNES_FLAG_C = 1     // Carry
SNES_FLAG_V = 2     // Overflow
SNES_FLAG_M = 4     // When E=0: Memory/accumulator operations 8-bit
SNES_FLAG_1 = 4     // When E=1: This bit is always set
SNES_FLAG_X = 8     // When E=0: Index registers 8-bit
SNES_FLAG_B = 8     // When E=1: Break (clear only on stack after IRQ/NMI)
SNES_FLAG_D = 0x10  // Decimal ADC/SBC mode
SNES_FLAG_I = 0x20  // Interrupt Disable
SNES_FLAG_Z = 0x40  // Zero result
SNES_FLAG_N = 0x80  // Negative result
SNES_FLAG_E = 0x100 // Emulation mode

SNES_FLAG_B1 = (SNES_FLAG_B | SNES_FLAG_1)
SNES_FLAG_MX = (SNES_FLAG_M | SNES_FLAG_X)
SNES_FLAG_NZ = (SNES_FLAG_N | SNES_FLAG_Z)
SNES_FLAG_NZC = (SNES_FLAG_NZ | SNES_FLAG_C)

REAL_SNES_FLAG_C = 1    // Carry
REAL_SNES_FLAG_Z = 2    // Zero result
REAL_SNES_FLAG_I = 4    // Interrupt Disable
REAL_SNES_FLAG_D = 8    // Decimal ADC/SBC mode
REAL_SNES_FLAG_X = 0x10 // When E=0: Index registers 8-bit
REAL_SNES_FLAG_B = 0x10 // When E=1: Break (clear only on stack after IRQ/NMI)
REAL_SNES_FLAG_M = 0x20 // When E=0: Memory/accumulator operations 8-bit
REAL_SNES_FLAG_1 = 0x20 // When E=1: This bit is always set
REAL_SNES_FLAG_V = 0x40 // Overflow
REAL_SNES_FLAG_N = 0x80 // Negative result
REAL_SNES_FLAG_E = 0x100    // Emulation mode

.macro  SET_FLAG flag
.if \flag >> 8
 orl $\flag,_P_W
.else
 orb $\flag,_P_B
.endif
.endm

.macro  CLR_FLAG flag
.if \flag >> 8
 andl $~\flag,_P_W
.else
 andb $~\flag,_P_B
.endif
.endm

.macro  CPL_FLAG flag
.if \flag >> 8
 xorl $\flag,_P_W
.else
 xorb $\flag,_P_B
.endif
.endm

.macro TEST_FLAG flag
.if \flag >> 8
 testl $\flag,_P_W  // ZF=!flag
.else
 testb $\flag,_P_B  // ZF=!flag
.endif
.endm

.macro JUMP_FLAG flag,wheretogo
 TEST_FLAG \flag
 jnz \wheretogo
.endm

.macro JUMP_NOT_FLAG flag,wheretogo
 TEST_FLAG \flag
 jz  \wheretogo
.endm

// Set 65816 flags to PC flags
// Can only set flags NZC - corrupts ah

.macro STORE_FLAGS flags
    lahf
    STORE_FLAGS_FROM \flags,%ah
.endm

// Set 65816 flags to PC flags
// Can only set flags NZC - corrupts loc
// Assumes loc has PC flags (as from lahf)
.macro STORE_FLAGS_FROM flags,loc
    andb $~(\flags),_P_B
    andb $\flags,\loc
    orb \loc,_P_B
.endm

// Set data to P
.macro FLAGS_TO data
.if \data >> 8
 movl $\data,_P
.else
 movb $\data,_P
.endif
.endm

// Load P to register _P_W
.macro LOAD_P
 movl _P,_P_W
.endm

// Save register _P_W to P
.macro SAVE_P
 movl _P_W,_P
.endm

// Load cycle counter to register _Cycles
.macro LOAD_CYCLES
 movl SNES_Cycles,_Cycles
.endm

// Save cycle counter to register _Cycles
.macro SAVE_CYCLES
 movl _Cycles,SNES_Cycles
.endm

// Load base pointer to CPU register set
.macro LOAD_BASE
 movl $CPU_Register_Base,_Base
.endm

// Load register _NativePC with pointer to code at PB:PC
.macro LOAD_PC
 movl _PBOffset,_NativePC
 addl _PC,_NativePC
.endm

// Get PC from register _NativePC
.macro GET_PC with
 movl _NativePC,\with
 subl _PBOffset,\with
.endm

// Save PC from register _NativePC
.macro SAVE_PC with
 GET_PC \with
 movl \with,_PC
.endm

// PC carry flag = 65816 carry flag (all others undefined)
.macro MAKE_CARRY
    rolb $8,_P_B
.endm

// PC carry flag = NOT 65816 carry flag (all others undefined)
.macro MAKE_NOT_CARRY
    rolb $8,_P_B
    cmc
.endm

// Set up the flags from PC flag format to 65816 flag format stored in %al
// Corrupts ebx
.macro  E0_SETUPFLAGS from=_P_B,to=%al
//.macro  Flags_Native_to_65c816_E0 from=_P_B,to=%al
    xorl %ebx,%ebx
    movb \from,%bl
    movb E0_Setup_Restore_Flags_Table(%ebx),\to
.endm

.macro  E1_SETUPFLAGS from=_P_B,to=%al
//.macro  Flags_Native_to_65c816_E1 from=_P_B,to=%al
    xorl %ebx,%ebx
    movb \from,%bl
    movb E1_Setup_Flags_Table(%ebx),\to
.endm

// Restore flags from SNES format in %al to PC format
// Corrupts eax
.macro  E0_RESTOREFLAGS to=_P_B
//.macro  Flags_65c816_to_Native_E0 to=_P_B
    andl $0xFF,%eax
    movb E0_Setup_Restore_Flags_Table(%eax),%al
    movb %al,\to
.endm

.macro  E1_RESTOREFLAGS to=_P_B
//.macro  Flags_65c816_to_Native_E1 to=_P_B
    andl $0xFF,%eax
    movb E1_Restore_Flags_Table(%eax),%al
    movb %al,\to
.endm

/*
 Set the current opcode execution and timing table pointers based
 on the M and X bits of flag register. This returns control to the
 execution loop, so it must be the last instruction
 Corrupts eax,ecx
*/

.macro  SET_TABLE_MX from=_P_B
    movb \from,%al
    andl $SNES_FLAG_MX,%eax // Mask out M & X flags
    JUMP_NOT_FLAG SNES_FLAG_X,7f
    movb %ah,_X+1           // Clear XH/YH if X flag set
    movb %ah,_Y+1
7:  movl CPU_OpTables(%eax),%eax
    movl %eax,OpTable
    jmp CPU_RETURN
.endm

/* Push/Pull macros assume eax contains value - corrupt ebx */

// Native mode - push byte (S--)
.macro  E0_PUSH_B data=%al
    movl _S,%ebx        // S only - bank always 0!
    cmpb $0x20,%bh
    jnb 0f
    movb \data,_WRAM(%ebx)
0:  decl %ebx           // Postdecrement S
    movw %bx,_S         // Set stack pointer
.endm

// Emulation mode - push byte (SL--)
.macro  E1_PUSH_B data=%al
    movl _S,%ebx        // S only - bank always 0!
    decb _S             // Postdecrement SL
    movb \data,_WRAM(%ebx)
.endm

// Native mode - pull byte (++S)
.macro  E0_PULL_B data=%al
    movl _S,%ebx        // S only - bank always 0!
    incw %bx            // Preincrement S
    movl %ebx,_S        // Set stack pointer
    cmpb $0x20,%bh
    jnb 0f
    movb _WRAM(%ebx),\data
0:
.endm

// Emulation mode - pull byte (++SL)
.macro  E1_PULL_B data=%al
    incb _S             // Preincrement SL
    movl _S,%ebx        // S only - bank always 0!
    movb _WRAM(%ebx),\data
.endm

// Native mode - push word (S--)
.macro  E0_PUSH_W datal=%al, datah=%ah
    movl _S,%ebx        // S only - bank always 0!
    cmpb $0x20,%bh
    jnb 0f
    movb \datah,_WRAM(%ebx)
0:  decw %bx            // Postdecrement S
    cmpb $0x20,%bh
    jnb 0f
    movb \datal,_WRAM(%ebx)
0:  decw %bx            // Postdecrement S
    movl %ebx,_S        // Set stack pointer
.endm

// Emulation mode - push word (SL--)
.macro  E1_PUSH_W datal=%al, datah=%ah
    movl _S,%ebx        // S only - bank always 0!
    movb \datah,_WRAM(%ebx)
    decb _S             // Postdecrement SL
    movl _S,%ebx        // S only - bank always 0!
    movb \datal,_WRAM(%ebx)
    decb _S             // Postdecrement SL
.endm

// Native mode - pull word (++S)
.macro  E0_PULL_W datal=%al, datah=%ah
    movl _S,%ebx        // S only - bank always 0!
    incw %bx            // Preincrement S
    cmpb $0x20,%bh
    jnb 0f
    movb _WRAM(%ebx),\datal
0:  incw %bx            // Preincrement S
    cmpb $0x20,%bh
    jnb 0f
    movb _WRAM(%ebx),\datah
0:  movl %ebx,_S        // Set stack pointer
.endm

// Emulation mode - pull word (++SL)
.macro  E1_PULL_W datal=%al, datah=%ah
    incb _S             // Preincrement SL
    movl _S,%ebx        // S only - bank always 0!
    movb _WRAM(%ebx),\datal
    incb _S             // Preincrement SL
    movl _S,%ebx        // S only - bank always 0!
    movb _WRAM(%ebx),\datah
.endm

// Requires current scanline to be in %eax
.macro CheckVIRQ
    movb NMITIMEN,%bl
    testb $0x30,%bl         // IRQ enabled?
    jz  0f
    testb $0x20,%bl         // V-IRQ enabled?
    jz  1f
    movl Current_Line_Timing,%edi
    cmpl %edi,VTIMEL
    jne 0f
1:  movl $0,%edx
    testb $0x10,%bl         // H-IRQ enabled?
    jz  1f
    movl HTimer,%edx
1:  movl FixedTrip,%edi
    cmpl %edx,%edi
    jb  2f
    movl %edx,_EventTrip
    movl $HIRQ_Event,Event_Handler
  //jmp 0f
2://movl Fixed_Handler,%edx
  //movl %edi,_EventTrip
  //movl %edx,Event_Handler
0:

.endm

#endif /* SNEeSe_CPU_i */
