/*
 SNEeSe SPC700 CPU emulation core
 Originally written by Savoury SnaX (Not quite sure if I like AT&T)
 Maintained/rewritten by Charles Bilyue'

 Compile under DJGPP (GNU-AS)
  NOTE - This must have .S (capital) or the # stuff wont be pre-processed!!!

 This file contains:
  CPU core info
  Reset
  Execution Loop
  Invalid Opcode Handler
  Flag format conversion tables
  Variable definitions (registers, interrupt vectors, etc.)
  CPU opcode emulation handlers
  CPU opcode handler table
  CPU opcode timing table

 CPU core info:
  Nearly all general registers are now used in SPC700 emulation:
   EAX,EBX are used by the memory mapper;
   CH is used to hold PSW register;
   EDX is used as CPU work register;
   EBP is used to hold cycle counter;
   ESI is used by the opcode fetcher;
   EDI is used as memory mapper work register.

    A register              - _A
    Y register              - _Y
    YA register pair        - _YA
    X register              - _X
    Stack pointer           - _SP
    Program Counter         - _PC
    Processor status word   - _PSW
       True x86 layout = |V|-|-|-|S|Z|-|A|-|-|-|C|
    True SPC700 layout =         |N|V|P|B|H|I|Z|C|
                   Using         |N|Z|P|H|B|I|V|C|

 SPC timers
  SPC700 timing is directly related to 65c816 timing, SPC gets 3 cycles for
  every 32 (21.48 MHz) 65c816 cycles. Since the timers run at 8KHz and 64KHz
  and the main chip runs at 2.048Mhz, the timers are clocked as follows:
    2.048MHz / 8KHz  = 256 cycles    (Timers 0 and 1)
    2.048MHz / 64KHz = 32  cycles    (Timer 2)

*/

#include "misc.i"
#include "SPC.i"

EXPORT_C SPC_start

// These are the bits for flag set/clr operations
SPC_FLAG_C = 1      // Carry
SPC_FLAG_V = 2      // Overflow
SPC_FLAG_I = 4      // Interrupt Disable
SPC_FLAG_B = 8      // Break
SPC_FLAG_H = 0x10   // Half-carry
SPC_FLAG_P = 0x20   // Page (direct page)
SPC_FLAG_Z = 0x40   // Zero result
SPC_FLAG_N = 0x80   // Negative result

SPC_FLAG_NZ = (SPC_FLAG_N | SPC_FLAG_Z)
SPC_FLAG_NZC = (SPC_FLAG_NZ | SPC_FLAG_C)
SPC_FLAG_NHZC = (SPC_FLAG_NZC | SPC_FLAG_H)

REAL_SPC_FLAG_C = 1     // Carry
REAL_SPC_FLAG_Z = 2     // Zero result
REAL_SPC_FLAG_I = 4     // Interrupt Disable
REAL_SPC_FLAG_H = 8     // Half-carry
REAL_SPC_FLAG_B = 0x10  // Break
REAL_SPC_FLAG_P = 0x20  // Page (direct page)
REAL_SPC_FLAG_V = 0x40  // Overflow
REAL_SPC_FLAG_N = 0x80  // Negative result

#define _PSW __SPC_PSW
#define _YA __SPC_YA
#define _A  __SPC_A
#define _Y  __SPC_Y
#define _X  __SPC_X
#define _SP __SPC_SP
#define _PC __SPC_PC

#define _PSW_B      _SPC700_PSW_B
#define _Base       _SPC700_Base
#define _Cycles     _SPC700_Cycles
#define _NativePC   _SPC700_NativePC

#define R_SPC_Code_Base     0x00(_Base)
#define R_PC                0x04(_Base)
#define R_YA                0x08(_Base)
#define R_A                 0x08(_Base)
#define R_Y                 0x09(_Base)
#define R_SPC_PAGE          0x0C(_Base)
#define R_SP                0x10(_Base)
#define R_PSW               0x14(_Base)
#define R_X                 0x15(_Base)
#define R_SPC_PORT0R        0x18(_Base)
#define R_SPC_PORT1R        0x19(_Base)
#define R_SPC_PORT2R        0x1A(_Base)
#define R_SPC_PORT3R        0x1B(_Base)
#define R_SPC_PORT0W        0x1C(_Base)
#define R_SPC_PORT1W        0x1D(_Base)
#define R_SPC_PORT2W        0x1E(_Base)
#define R_SPC_PORT3W        0x1F(_Base)
#ifdef DEBUG
#define R_SPC_TEMP_ADD      0x20(_Base)
#endif

// Load PSW to register _PSW_B
.macro LOAD_PSW
 movb _PSW,_PSW_B
.endm

// Save register _PSW_B to PSW
.macro SAVE_PSW
 movb _PSW_B,_PSW
.endm

// Load cycle counter to register _Cycles
.macro LOAD_CYCLES
  movl _TotalCycles,_Cycles
.endm

// Save register _Cycles to cycle counter
.macro SAVE_CYCLES
  movl _Cycles,_TotalCycles
.endm

// Load base pointer to CPU register set
.macro LOAD_BASE
  movl $SPC_Register_Base,_Base
.endm

// Load register _NativePC with pointer to code at PC
.macro LOAD_PC
 movl SPC_Code_Base,_NativePC
 addl _PC,_NativePC
.endm

// Get PC from register _NativePC
.macro GET_PC with
 movl _NativePC,\with
 subl SPC_Code_Base,\with
.endm

// Save PC from register _NativePC
.macro SAVE_PC with
 GET_PC \with
 movl \with,_PC
.endm

// Set up the flags from PC flag format to SPC flag format stored in %al
// Corrupts eax
.macro  SETUPFLAGS_SPC from=_PSW_B,to=%al
//.macro  Flags_Native_to_SPC from=_PSW_B,to=%al
    xorl %eax,%eax
    movb \from,%al
    movb SPC_Setup_Restore_Flags_Table(%eax),\to
.endm

// Restore flags from SPC format in %al to PC format
// Corrupts eax
.macro  RESTOREFLAGS_SPC to=_PSW_B
//.macro  Flags_SPC_to_Native to=_PSW_B
    andl $0xFF,%eax
    testb $REAL_SPC_FLAG_P,%al
    movb SPC_Setup_Restore_Flags_Table(%eax),%al
    setnz SPC_PAGE+1        // Setup direct page bit
    movb %al,\to
.endm

/* SPC MEMORY MAPPER IS PLACED HERE (ITS SIMPLER THAN THE CPU ONE!) */

/* bx - contains the actual address, al is where the info should be stored, edi is free */
/* NB bx is not corrupted! edi is corrupted! */
/* NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc! */

ALIGNC
SPC_READ_MAPPER:
    andl $0xFFFF,%ebx
    testb %bh,%bh
    jz  SPC_READ_ZERO_PAGE

    cmpl $0xFFC0,%ebx
    jae SPC_READ_RAM_ROM
SPC_READ_RAM:
    movb _SPCRAM(%ebx),%al
    ret

ALIGNC
SPC_READ_RAM_ROM:
    movb SPC_ROM_CODE - 0xFFC0(%ebx),%al
SPC_FFC0_Address=.-4
    ret

ALIGNC
// This code should be copied into the top of the address space
SPC_ROM_CODE:
    .byte	0xCD,0xEF,0xBD,0xE8,0x00,0xC6,0x1D,0xD0
    .byte	0xFC,0x8F,0xAA,0xF4,0x8F,0xBB,0xF5,0x78
    .byte	0xCC,0xF4,0xD0,0xFB,0x2F,0x19,0xEB,0xF4
    .byte	0xD0,0xFC,0x7E,0xF4,0xD0,0x0B,0xE4,0xF5
    .byte	0xCB,0xF4,0xD7,0x00,0xFC,0xD0,0xF3,0xAB
    .byte	0x01,0x10,0xEF,0x7E,0xF4,0x10,0xEB,0xBA
    .byte	0xF6,0xDA,0x00,0xBA,0xF4,0xC4,0xF4,0xDD
    .byte	0x5D,0xD0,0xDB,0x1F,0x00,0x00,0xC0,0xFF

ALIGNC
SPC_READ_ZERO_PAGE:
    cmpb $0xF0,%bl
    jb  SPC_READ_RAM

SPC_READ_FUNC:
    jmpl Read_Func_Map - 0xF0 * 4(,%ebx,4)

.balign 4
Read_Func_Map:              // Mappings for SPC Registers
    .long	SPC_READ_INVALID
    .long	SPC_READ_INVALID
    .long	SPC_READ_DSP_ADDR
    .long	SPC_READ_DSP_DATA
    .long   SPC_READ_PORT0R
    .long   SPC_READ_PORT1R
    .long   SPC_READ_PORT2R
    .long   SPC_READ_PORT3R
    .long	SPC_READ_INVALID
    .long	SPC_READ_INVALID
    .long	SPC_READ_INVALID
    .long	SPC_READ_INVALID
    .long	SPC_READ_INVALID
    .long	SPC_READ_COUNTER_0
    .long	SPC_READ_COUNTER_1
    .long	SPC_READ_COUNTER_2

ALIGNC
SPC_READ_INVALID:
    movb $0xFF,%al          // v0.15
#ifdef DEBUG
//  andl $0xFFFF,%ebx
    movl %ebx,_Map_Address  // Set up Map Address so message works!
    movb %al,_Map_Byte      // Set up Map Byte so message works

    pusha
    call _InvalidSPCHWRead  // Display read from invalid HW warning
    popa
#endif
    ret

/*   --------   */

SPC_WRITE_MAPPER:
    andl $0xFFFF,%ebx
    testb %bh,%bh
    jz  SPC_WRITE_ZERO_PAGE

SPC_WRITE_RAM:
    movb %al,_SPCRAM(%ebx)
    ret

ALIGNC
SPC_WRITE_ZERO_PAGE:
    cmpb $0xF0,%bl
    jb  SPC_WRITE_RAM

SPC_WRITE_FUNC:
    jmpl Write_Func_Map - 0xF0 * 4(,%ebx,4)

.balign 4
Write_Func_Map:             // Mappings for SPC Registers
    .long	SPC_WRITE_INVALID
    .long	SPC_WRITE_CTRL
    .long	SPC_WRITE_DSP_ADDR
    .long	SPC_WRITE_DSP_DATA
    .long   SPC_WRITE_PORT0W
    .long   SPC_WRITE_PORT1W
    .long   SPC_WRITE_PORT2W
    .long   SPC_WRITE_PORT3W
    .long	SPC_WRITE_INVALID
    .long	SPC_WRITE_INVALID
    .long	SPC_WRITE_TIMER_0
    .long	SPC_WRITE_TIMER_1
    .long	SPC_WRITE_TIMER_2
    .long	SPC_WRITE_INVALID
    .long	SPC_WRITE_INVALID
    .long	SPC_WRITE_INVALID

SPC_WRITE_INVALID:
#ifdef DEBUG
//  andl $0xFFFF,%ebx
    movl %ebx,_Map_Address      // Set up Map Address so message works!
    movb %al,_Map_Byte          // Set up Map Byte so message works

    pusha
    call _InvalidSPCHWWrite // Display write to invalid HW warning
    popa
#endif
    ret

// GET_BYTE & GET_WORD now assume ebx contains the read address and 
// eax the place to store value also, corrupts edi

.macro  GET_BYTE_SPC
    call SPC_READ_MAPPER
.endm

.macro  GET_WORD_SPC
    GET_BYTE_SPC
    movb %al,%ah
    incw %bx
    GET_BYTE_SPC
    xchgb %al,%ah
.endm

// SET_BYTE & SET_WORD now assume ebx contains the write address and 
// eax the value to write, corrupts edi

.macro  SET_BYTE_SPC
    call SPC_WRITE_MAPPER
.endm

.macro  SET_WORD_SPC
    SET_BYTE_SPC
    movb %ah,%al
    incw %bx
    SET_BYTE_SPC
.endm

// --- Ease up on the finger cramps ;-)

.macro SET_FLAG_SPC flag
    orb $\flag,_PSW_B
.endm

.macro CLR_FLAG_SPC flag
    andb $~(\flag),_PSW_B
.endm

.macro CPL_FLAG_SPC flag
    xorb $\flag,_PSW_B
.endm

.macro TEST_FLAG_SPC flag      // Set ZF=!flag
    testb $\flag,_PSW_B
.endm

.macro JUMP_FLAG_SPC flag,wheretogo
    TEST_FLAG_SPC \flag
    jnz \wheretogo
.endm

.macro JUMP_NOT_FLAG_SPC flag,wheretogo
    TEST_FLAG_SPC \flag
    jz  \wheretogo
.endm

// Push / Pop Macros assume eax contains value - corrupt ebx,edi
.macro PUSH_B           // Push Byte (SP--)
    movl R_SP,%ebx
    movb %al,_SPCRAM(%ebx)  // Store data on stack
    decl %ebx
    movb %bl,R_SP       // Decrement S (Byte)
.endm

.macro POP_B            // Pop Byte (++SP)
    movl R_SP,%ebx
    incb %bl
    movb %bl,R_SP
    movb _SPCRAM(%ebx),%al  // Fetch data from stack
.endm

.macro PUSH_W            // Push Word (SP--)
    movl R_SP,%ebx
    movb %ah,_SPCRAM(%ebx)      // Store data on stack
    movb %al,_SPCRAM-1(%ebx)    // Store data on stack
    subb $2,%bl
    movb %bl,R_SP       // Postdecrement SP
.endm

.macro POP_W            // Pop Word (++SP)
    movl R_SP,%ebx
    addb $2,%bl         // Preincrement SP
    movb %bl,R_SP
    movb _SPCRAM(%ebx),%ah      // Fetch data from stack
    movb _SPCRAM-1(%ebx),%al    // Fetch data from stack
.endm

// Set SPC700 flags to PC flags
// Can only set flags NHZC - corrupts ah

.macro STORE_FLAGS_SPC flags
    lahf
    STORE_FLAGS_SPC_FROM \flags,%ah
.endm

// Set SPC700 flags to PC flags
// Can only set flags NHZC - corrupts loc
// Assumes loc has PC flags (as from lahf)
.macro STORE_FLAGS_SPC_FROM flags,loc
    andb $(~\flags),_PSW_B
    andb $(\flags),\loc
    orb \loc,_PSW_B
.endm

// Set PC Flags To SNES Flags (only sets required flags all others cleared!) corrupts eax

.macro	MAKE_FLAGS_SPC flags
    movb _PSW_B,%ah
    andb $\flags,%ah    // Flags ready
    sahf
.endm

// Set PC Flags To SNES Flags (Inverts bit (e.g. C=1 means clear carry)) corrupts eax

.macro	MAKE_FLAGS_SPC_INV flags
    movb _PSW_B,%ah
    xorb $0xFF,%ah
    andb $\flags,%ah    // Flags ready
    sahf
.endm

// PC carry flag = SPC700 carry flag (all others undefined)
.macro MAKE_CARRY
    rolb $8,_PSW_B
.endm

// PC carry flag = NOT SPC700 carry flag (all others undefined)
.macro MAKE_NOT_CARRY
    rolb $8,_PSW_B
    cmc
.endm

.data
ALIGND
.globl SPCOpTable
SPCOpTable:
/* 00 */ .long SPC_NOP          ,SPC_TCALL_0      ,SPC_SET1         ,SPC_BBS 
/* 04 */ .long SPC_OR_A_dp      ,SPC_OR_A_abs     ,SPC_OR_A_OXO     ,SPC_OR_A_OOdp_XOO
/* 08 */ .long SPC_OR_A_IM      ,SPC_OR_dp_dp     ,SPC_OR1          ,SPC_ASL_dp
/* 0C */ .long SPC_ASL_abs      ,SPC_PUSH_PSW     ,SPC_TSET1        ,SPC_INVALID
/* 10 */ .long SPC_BPL          ,SPC_TCALL_1      ,SPC_CLR1         ,SPC_BBC 
/* 14 */ .long SPC_OR_A_Odp_XO  ,SPC_OR_A_Oabs_XO ,SPC_OR_A_Oabs_YO ,SPC_OR_A_OOdpO_YO
/* 18 */ .long SPC_OR_dp_IM     ,SPC_OR_OXO_OYO   ,SPC_DECW_dp      ,SPC_ASL_Odp_XO
/* 1C */ .long SPC_ASL_A        ,SPC_DEC_X        ,SPC_CMP_X_abs    ,SPC_JMP_Oabs_XO
/* 20 */ .long SPC_CLRP         ,SPC_TCALL_2      ,SPC_SET1         ,SPC_BBS 
/* 24 */ .long SPC_AND_A_dp     ,SPC_AND_A_abs    ,SPC_AND_A_OXO    ,SPC_AND_A_OOdp_XOO
/* 28 */ .long SPC_AND_A_IM     ,SPC_AND_dp_dp    ,SPC_OR1C         ,SPC_ROL_dp
/* 2C */ .long SPC_ROL_abs      ,SPC_PUSH_A       ,SPC_CBNE_dp      ,SPC_BRA
/* 30 */ .long SPC_BMI          ,SPC_TCALL_3      ,SPC_CLR1         ,SPC_BBC 
/* 34 */ .long SPC_AND_A_Odp_XO ,SPC_AND_A_Oabs_XO,SPC_AND_A_Oabs_YO,SPC_AND_A_OOdpO_YO
/* 38 */ .long SPC_AND_dp_IM    ,SPC_AND_OXO_OYO  ,SPC_INCW_dp      ,SPC_ROL_Odp_XO
/* 3C */ .long SPC_ROL_A        ,SPC_INC_X        ,SPC_CMP_X_dp     ,SPC_CALL
/* 40 */ .long SPC_SETP         ,SPC_TCALL_4      ,SPC_SET1         ,SPC_BBS 
/* 44 */ .long SPC_EOR_A_dp     ,SPC_EOR_A_abs    ,SPC_EOR_A_OXO    ,SPC_EOR_A_OOdp_XOO
/* 48 */ .long SPC_EOR_A_IM     ,SPC_EOR_dp_dp    ,SPC_AND1         ,SPC_LSR_dp
/* 4C */ .long SPC_LSR_abs      ,SPC_PUSH_X       ,SPC_TCLR1        ,SPC_PCALL
/* 50 */ .long SPC_BVC          ,SPC_TCALL_5      ,SPC_CLR1         ,SPC_BBC 
/* 54 */ .long SPC_EOR_A_Odp_XO ,SPC_EOR_A_Oabs_XO,SPC_EOR_A_Oabs_YO,SPC_EOR_A_OOdpO_YO
/* 58 */ .long SPC_EOR_dp_IM    ,SPC_EOR_OXO_OYO  ,SPC_CMPW_YA_dp   ,SPC_LSR_Odp_XO
/* 5C */ .long SPC_LSR_A        ,SPC_MOV_X_A      ,SPC_CMP_Y_abs    ,SPC_JMP_abs
/* 60 */ .long SPC_CLRC         ,SPC_TCALL_6      ,SPC_SET1         ,SPC_BBS 
/* 64 */ .long SPC_CMP_A_dp     ,SPC_CMP_A_abs    ,SPC_CMP_A_OXO    ,SPC_CMP_A_OOdp_XOO
/* 68 */ .long SPC_CMP_A_IM     ,SPC_CMP_dp_dp    ,SPC_AND1C        ,SPC_ROR_dp
/* 6C */ .long SPC_ROR_abs      ,SPC_PUSH_Y       ,SPC_DBNZ_dp      ,SPC_RET
/* 70 */ .long SPC_BVS          ,SPC_TCALL_7      ,SPC_CLR1         ,SPC_BBC 
/* 74 */ .long SPC_CMP_A_Odp_XO ,SPC_CMP_A_Oabs_XO,SPC_CMP_A_Oabs_YO,SPC_CMP_A_OOdpO_YO
/* 78 */ .long SPC_CMP_dp_IM    ,SPC_CMP_OXO_OYO  ,SPC_ADDW_YA_dp   ,SPC_ROR_Odp_XO
/* 7C */ .long SPC_ROR_A        ,SPC_MOV_A_X      ,SPC_CMP_Y_dp     ,SPC_INVALID
/* 80 */ .long SPC_SETC         ,SPC_TCALL_8      ,SPC_SET1         ,SPC_BBS 
/* 84 */ .long SPC_ADC_A_dp     ,SPC_ADC_A_abs    ,SPC_ADC_A_OXO    ,SPC_ADC_A_OOdp_XOO
/* 88 */ .long SPC_ADC_A_IM     ,SPC_ADC_dp_dp    ,SPC_EOR1         ,SPC_DEC_dp
/* 8C */ .long SPC_DEC_abs      ,SPC_MOV_Y_IM     ,SPC_POP_PSW      ,SPC_MOV_dp_IM
/* 90 */ .long SPC_BCC          ,SPC_TCALL_9      ,SPC_CLR1         ,SPC_BBC 
/* 94 */ .long SPC_ADC_A_Odp_XO ,SPC_ADC_A_Oabs_XO,SPC_ADC_A_Oabs_YO,SPC_ADC_A_OOdpO_YO
/* 98 */ .long SPC_ADC_dp_IM    ,SPC_ADC_OXO_OYO  ,SPC_SUBW_YA_dp   ,SPC_DEC_Odp_XO
/* 9C */ .long SPC_DEC_A        ,SPC_MOV_X_SP     ,SPC_DIV          ,SPC_XCN
/* A0 */ .long SPC_EI           ,SPC_TCALL_10     ,SPC_SET1         ,SPC_BBS 
/* A4 */ .long SPC_SBC_A_dp     ,SPC_SBC_A_abs    ,SPC_SBC_A_OXO    ,SPC_SBC_A_OOdp_XOO
/* A8 */ .long SPC_SBC_A_IM     ,SPC_SBC_dp_dp    ,SPC_MOV1_C_      ,SPC_INC_dp
/* AC */ .long SPC_INC_abs      ,SPC_CMP_Y_IM     ,SPC_POP_A        ,SPC_MOV_OXOInc_A
/* B0 */ .long SPC_BCS          ,SPC_TCALL_11     ,SPC_CLR1         ,SPC_BBC 
/* B4 */ .long SPC_SBC_A_Odp_XO ,SPC_SBC_A_Oabs_XO,SPC_SBC_A_Oabs_YO,SPC_SBC_A_OOdpO_YO
/* B8 */ .long SPC_SBC_dp_IM    ,SPC_SBC_OXO_OYO  ,SPC_MOV_YA_dp    ,SPC_INC_Odp_XO
/* BC */ .long SPC_INC_A        ,SPC_MOV_SP_X     ,SPC_INVALID      ,SPC_MOV_A_OXOInc
/* C0 */ .long SPC_DI           ,SPC_TCALL_12     ,SPC_SET1         ,SPC_BBS 
/* C4 */ .long SPC_MOV_dp_A     ,SPC_MOV_abs_A    ,SPC_MOV_OXO_A    ,SPC_MOV_OOdp_XOO_A
/* C8 */ .long SPC_CMP_X_IM     ,SPC_MOV_abs_X    ,SPC_MOV1__C      ,SPC_MOV_dp_Y
/* CC */ .long SPC_MOV_abs_Y    ,SPC_MOV_X_IM     ,SPC_POP_X        ,SPC_MUL
/* D0 */ .long SPC_BNE          ,SPC_TCALL_13     ,SPC_CLR1         ,SPC_BBC 
/* D4 */ .long SPC_MOV_Odp_XO_A ,SPC_MOV_Oabs_XO_A,SPC_MOV_Oabs_YO_A,SPC_MOV_OOdpO_YO_A
/* D8 */ .long SPC_MOV_dp_X     ,SPC_MOV_Odp_YO_X ,SPC_MOV_dp_YA    ,SPC_MOV_Odp_XO_Y
/* DC */ .long SPC_DEC_Y        ,SPC_MOV_A_Y      ,SPC_CBNE_Odp_XO  ,SPC_INVALID
/* E0 */ .long SPC_CLRV         ,SPC_TCALL_14     ,SPC_SET1         ,SPC_BBS 
/* E4 */ .long SPC_MOV_A_dp     ,SPC_MOV_A_abs    ,SPC_MOV_A_OXO    ,SPC_MOV_A_OOdp_XOO
/* E8 */ .long SPC_MOV_A_IM     ,SPC_MOV_X_abs    ,SPC_NOT1         ,SPC_MOV_Y_dp
/* EC */ .long SPC_MOV_Y_abs    ,SPC_NOTC         ,SPC_POP_Y        ,SPC_INVALID /*SPC_SLEEP*/
/* F0 */ .long SPC_BEQ          ,SPC_TCALL_15     ,SPC_CLR1         ,SPC_BBC 
/* F4 */ .long SPC_MOV_A_Odp_XO ,SPC_MOV_A_Oabs_XO,SPC_MOV_A_Oabs_YO,SPC_MOV_A_OOdpO_YO
/* F8 */ .long SPC_MOV_X_dp     ,SPC_MOV_X_Odp_YO ,SPC_MOV_dp_dp    ,SPC_MOV_Y_Odp_XO
/* FC */ .long SPC_INC_Y        ,SPC_MOV_Y_A      ,SPC_DBNZ_Y       ,SPC_INVALID /*SPC_STOP*/

// This holds the base instruction timings in cycles
SPCCycleTable:
/* 00 */ .byte 2,8,4,5,3,4,3,6,2,6,5,4,5,4,6,8
/* 10 */ .byte 2,8,4,5,4,5,5,6,5,5,6,5,2,2,4,6
/* 20 */ .byte 2,8,4,5,3,4,3,6,2,6,5,4,5,4,5,4
/* 30 */ .byte 2,8,4,5,4,5,5,6,5,5,6,5,2,2,3,8
/* 40 */ .byte 2,8,4,5,3,4,3,6,2,6,4,4,5,4,6,6
/* 50 */ .byte 2,8,4,5,4,5,5,6,5,5,4,5,2,2,4,3
/* 60 */ .byte 2,8,4,5,3,4,3,6,2,6,4,4,5,4,5,5
/* 70 */ .byte 2,8,4,5,4,5,5,6,5,5,5,5,2,2,3,6
/* 80 */ .byte 2,8,4,5,3,4,3,6,2,6,5,4,5,2,4,5
/* 90 */ .byte 2,8,4,5,4,5,5,6,5,5,5,5,2,2,12,5
/* A0 */ .byte 3,8,4,5,3,4,3,6,2,6,4,4,5,2,4,4
/* B0 */ .byte 2,8,4,5,4,5,5,6,5,5,5,5,2,2,3,4
/* C0 */ .byte 3,8,4,5,4,5,4,7,2,5,6,4,5,2,4,9
/* D0 */ .byte 2,8,4,5,5,6,6,7,4,5,4,5,2,2,6,3
/* E0 */ .byte 2,8,4,5,3,4,3,6,2,4,5,3,4,3,4,3
/* F0 */ .byte 2,8,4,5,4,5,5,6,3,4,5,4,2,2,4,3

#include "SPCflags.S"

ALIGND
offset_to_bit:  .byte 0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80
offset_to_not:  .byte 0xFE,0xFD,0xFB,0xF7,0xEF,0xDF,0xBF,0x7F

.bss
ALIGND
SPC_Register_Base:

.globl _PC,_SP,_YA,_A,_Y,_PSW,_X
SPC_Code_Base:  skipl
_PC:skipl
_YA:
_A: skipb
_Y: skipb
    skipw
SPC_PAGE:       skipl
_SP:skipl

_PSW:   skipb   // Processor status word
_X: skipb

.balign 4
.globl _SPC_PORT0R,_SPC_PORT1R,_SPC_PORT2R,_SPC_PORT3R
_SPC_PORT0R:skipb
_SPC_PORT1R:skipb
_SPC_PORT2R:skipb
_SPC_PORT3R:skipb
.globl _SPC_PORT0W,_SPC_PORT1W,_SPC_PORT2W,_SPC_PORT3W
_SPC_PORT0W:skipb
_SPC_PORT1W:skipb
_SPC_PORT2W:skipb
_SPC_PORT3W:skipb

#ifdef DEBUG
SPC_TEMP_ADD:   skipl
#endif

.text
ALIGNC
SNES_R2140_SPC: // SPCI00
    Execute_SPC
    movb _SPC_PORT0W,%al
    ret

ALIGNC
SNES_R2141_SPC: // SPCI01
    Execute_SPC
    movb _SPC_PORT1W,%al
    ret

ALIGNC
SNES_R2142_SPC: // SPCI02
    Execute_SPC
    movb _SPC_PORT2W,%al
    ret

ALIGNC
SNES_R2143_SPC: // SPCI03
    Execute_SPC
    movb _SPC_PORT3W,%al
    ret

ALIGNC
SNES_W2140_SPC: // SPCI00
    Execute_SPC
    movb %al,_SPC_PORT0R
    ret

ALIGNC
SNES_W2141_SPC: // SPCI01
    Execute_SPC
    movb %al,_SPC_PORT1R
    ret

ALIGNC
SNES_W2142_SPC: // SPCI02
    Execute_SPC
    movb %al,_SPC_PORT2R
    ret

ALIGNC
SNES_W2143_SPC: // SPCI03
    Execute_SPC
    movb %al,_SPC_PORT3R
    ret

ALIGNC
SPC_SET1:
    shrl $5,%eax
    movl SPC_PAGE,%ebx
    movb 1(_NativePC),%bl     
    addl $2,_NativePC
    movb offset_to_bit(%eax),%ah
    GET_BYTE_SPC     
    orb %ah,%al
    SET_BYTE_SPC
    jmp SPC_BACK

ALIGNC
SPC_CLR1:
    shrl $5,%eax
    movl SPC_PAGE,%ebx
    movb 1(_NativePC),%bl     
    addl $2,_NativePC
    movb offset_to_not(%eax),%ah
    GET_BYTE_SPC     
    andb %ah,%al
    SET_BYTE_SPC
    jmp SPC_BACK

ALIGNC
SPC_BBS:
    shrl $5,%eax
    movl SPC_PAGE,%ebx
    movb 1(_NativePC),%bl     
    movb offset_to_bit(%eax),%ah
    GET_BYTE_SPC         
    testb %ah,%al
    jz  9f               
    movsbl 2(_NativePC),%eax  
    addl %eax,_NativePC
    addl $2,_Cycles     // branch taken
9:  addl $3,_NativePC
    jmp SPC_BACK

ALIGNC
SPC_BBC:
    shrl $5,%eax
    movl SPC_PAGE,%ebx
    movb 1(_NativePC),%bl     
    movb offset_to_bit(%eax),%ah
    GET_BYTE_SPC         
    testb %ah,%al
    jnz 9f               
    movsbl 2(_NativePC),%eax  
    addl %eax,_NativePC
    addl $2,_Cycles     // branch taken
9:  addl $3,_NativePC
    jmp SPC_BACK

ALIGNC
EXPORT_C Make_SPC
    pusha
    movl $SNES_R2140_SPC,%eax
    movl %eax,Read_Map_21+0x40*4
    movl $SNES_W2140_SPC,%eax
    movl %eax,Write_Map_21+0x40*4
    movl $SNES_R2141_SPC,%eax
    movl %eax,Read_Map_21+0x41*4
    movl $SNES_W2141_SPC,%eax
    movl %eax,Write_Map_21+0x41*4
    movl $SNES_R2142_SPC,%eax
    movl %eax,Read_Map_21+0x42*4
    movl $SNES_W2142_SPC,%eax
    movl %eax,Write_Map_21+0x42*4
    movl $SNES_R2143_SPC,%eax
    movl %eax,Read_Map_21+0x43*4
    movl $SNES_W2143_SPC,%eax
    movl %eax,Write_Map_21+0x43*4
    popa
    ret

ALIGNC
SPC_INVALID:
    movb %al,_Map_Byte      // al contains opcode!

    xorl %eax,%eax
    decw _PC                // Adjust address to correct for pre-increment
    movl _PC,%eax
    movl %eax,_Map_Address  // this just sets the error output up correctly!

    movb _PSW_B,%al
    movb %al,_PSW

    jmp _InvalidSPCOpcode      // This exits.. avoids conflict with other things!

ALIGNC
EXPORT_C Reset_SPC
    pusha

    cld

    // This sets up the ROM code, done before registers as it
    // contains the reset address for the ROM code (0xFFC0)
    movl $SPC_ROM_CODE,%esi
    movl $_SPCRAM+0xFFC0,%edi
    movl $64/4,%ecx
    rep; movsl

    // Get ROM reset vector and setup Program Counter
    movzwl SPC_ROM_CODE+(0xFFFE-0xFFC0),%eax
    movl %eax,_PC

    movl SNES_Cycles,%eax
    movl %eax,SPC_last_cycles

    // Reset the sound DSP registers
    xorl %eax,%eax
    movl $_SPC_DSP,%edi
    movl $128/4,%ecx
    rep; stosl

    movl %eax,_SPC_Cycles   // Clear Cycle Count
    movl %eax,_TotalCycles
    movl %eax,SPC_PAGE      // Used to save looking up P flag for Direct page stuff!
    movl $0x01FF,_SP        // Reset registers
    movl %eax,_YA
    movb %al,_X
    movb %al,_PSW           // Clear Flags Register

    movb $0x80,_SPC_CTRL
    movl $SPC_ROM_CODE-0xFFC0,SPC_FFC0_Address
    movl $SPC_ROM_CODE-0xFFC0,SPC_Code_Base

    // Reset timers
    movb %al,_SPC_CNT0
    movb %al,_SPC_CNT1
    movb %al,_SPC_CNT2
    movl $256,_SPC_TIMT0
    movl $256,_SPC_TIMT1
    movl $256,_SPC_TIMT2
    movl %eax,_SPC_TIMER0
    movl %eax,_SPC_TIMER1
    movl %eax,_SPC_TIMER2
    movl %eax,_CycleLatch0
    movl %eax,_CycleLatch1
    movl %eax,_CycleLatch2

    // Reset SPC700 output ports
    movb %al,_SPC_PORT0W
    movb %al,_SPC_PORT1W
    movb %al,_SPC_PORT2W
    movb %al,_SPC_PORT3W

    decl %eax
    // Reset SPC700 input ports
    movb %al,_SPC_PORT0R
    movb %al,_SPC_PORT1R
    movb %al,_SPC_PORT2R
    movb %al,_SPC_PORT3R

    // Reset sound DSP port address
    movb %al,_SPC_DSP_ADDR

    popa
    ret

SPC_SHOW_REGISTERS:
    pusha
//  movb _PSW_B,%al
//  movb %al,_PSW
    call _DisplaySPC
    popa
    ret

ALIGNC
.globl SPC_START
SPC_START:
  movb CPU_Executing,%al
  pushl %eax
  movb $0,CPU_Executing

  LOAD_CYCLES
  LOAD_PSW
  LOAD_PC
  LOAD_BASE
  xorl %eax,%eax
  jmp SPC_LOOP

ALIGNC
SPC_BACK:
//cmpl $SPC_Register_Base,_Base
//jne 0b

#ifdef DEBUG
//movl SPC_TEMP_ADD,%ebx
//movl %ebx,_OLD_SPC_ADDRESS
#endif
  cmpl $0x80000000,_Cycles
.globl _SPC_Cycles
_SPC_Cycles=.-4         // Number of cycles to execute for SPC
  ja  SPC_OUT           // Do another instruction if cycles left

SPC_LOOP:
//  movl _PC,%ebx         // PC now setup
//  movl SPC_Code_Base,_NativePC
//  addl %ebx,_NativePC
  xorl %eax,%eax
#ifdef DEBUG
//movl %ebx,SPC_TEMP_ADD
#endif
  movb (_NativePC),%al  // Fetch opcode
  movl %eax,%ebx
//  incl %ebx
  movb SPCCycleTable(%eax),%bl
  addl %ebx,_Cycles             // Update cycle counter
  jmpl SPCOpTable(,%eax,4)      // jmp to opcode handler

ALIGNC
SPC_OUT:
  SAVE_PC %ebx
  SAVE_PSW
  SAVE_CYCLES   // Set cycle counter

  popl %eax
  movb %al,CPU_Executing
  ret                   // Return to CPU emulation

#include "SPCaddr.S"    // Include addressing mode macros
#include "SPCmacro.S"   // Include instruction macros

#include "SPC0.S"       // Include opcodes 0x00-0x0F
#include "SPC1.S"       // Include opcodes 0x10-0x1F
#include "SPC2.S"       // Include opcodes 0x20-0x2F
#include "SPC3.S"       // Include opcodes 0x30-0x3F
#include "SPC4.S"       // Include opcodes 0x40-0x4F
#include "SPC5.S"       // Include opcodes 0x50-0x5F
#include "SPC6.S"       // Include opcodes 0x60-0x6F
#include "SPC7.S"       // Include opcodes 0x70-0x7F
#include "SPC8.S"       // Include opcodes 0x80-0x8F
#include "SPC9.S"       // Include opcodes 0x90-0x9F
#include "SPCA.S"       // Include opcodes 0xA0-0xAF
#include "SPCB.S"       // Include opcodes 0xB0-0xBF
#include "SPCC.S"       // Include opcodes 0xC0-0xCF
#include "SPCD.S"       // Include opcodes 0xD0-0xDF
#include "SPCE.S"       // Include opcodes 0xE0-0xEF
#include "SPCF.S"       // Include opcodes 0xF0-0xFF

SPC_READ_DSP_ADDR:
    movb _SPC_DSP_ADDR,%al
    ret

SPC_READ_DSP_DATA:
    pusha
    call _SPC_READ_DSP
    popa
    pushl %ebx
    movb _SPC_DSP_ADDR,%bl
    andl $0x7F,%ebx
    movb _SPC_DSP(,%ebx),%al    // read from DSP register
    popl %ebx
    ret

SPC_READ_PORT0R:
    movb _SPC_PORT0R,%al
    ret
SPC_READ_PORT1R:
    movb _SPC_PORT1R,%al
    ret
SPC_READ_PORT2R:
    movb _SPC_PORT2R,%al
    ret
SPC_READ_PORT3R:
    movb _SPC_PORT3R,%al
    ret

// WOOPS... TIMER registers are write only, the actual timer clock is internal not accessible!

// COUNTERS ARE 4 BIT, upon read they reset to 0 status

SPC_READ_COUNTER_0:
    pusha
    SAVE_CYCLES   // Set cycle counter
    call _Update_SPC_Timer_0
    popa
    movb _SPC_CNT0,%al
    movb $0,_SPC_CNT0
    ret

SPC_READ_COUNTER_1:
    pusha
    SAVE_CYCLES   // Set cycle counter
    call _Update_SPC_Timer_1
    popa
    movb _SPC_CNT1,%al
    movb $0,_SPC_CNT1
    ret

SPC_READ_COUNTER_2:
    pusha
    SAVE_CYCLES   // Set cycle counter
    call _Update_SPC_Timer_2
    popa
    movb _SPC_CNT2,%al
    movb $0,_SPC_CNT2
    ret

// | ---- | ---- | PC32 | PC10 | ---- | ST2  | ST1  | ST0  |
SPC_WRITE_CTRL:
    pushl %eax
    movb $0,%ah
    testb %al,%al       // New for 0.25 - read hidden RAM
    movl $_SPCRAM,%edi
    jns 1f
    movl $SPC_ROM_CODE-0xFFC0,%edi
1:  movl %edi,SPC_FFC0_Address
    testb $0x10,%al     // Reset ports 0/1 to 00 if set
    jz  0f
    movb %ah,_SPC_PORT0R    // Ports read by SPC should be reset! 
    movb %ah,_SPC_PORT1R    // Thanks to Butcha for fix!
0:  testb $0x20,%al     // Reset ports 2/3 to 00 if set
    jz  0f
    movb %ah,_SPC_PORT2R
    movb %ah,_SPC_PORT3R
0:  movl _Cycles,%edi
    testb $4,_SPC_CTRL
    jnz 0f
    testb $4,%al
    jz  0f
    movb $0,_SPC_CNT2
    movl $0,_SPC_TIMER2
    movl %edi,_CycleLatch2
0:  testb $2,_SPC_CTRL
    jnz 0f
    testb $2,%al
    jz  0f
    movb $0,_SPC_CNT1
    movl $0,_SPC_TIMER1
    movl %edi,_CycleLatch1
0:  testb $1,_SPC_CTRL
    jnz 0f
    testb $1,%al
    jz  0f
    movb $0,_SPC_CNT0
    movl $0,_SPC_TIMER0
    movl %edi,_CycleLatch0
0:  popl %eax
    movb %al,_SPC_CTRL
    ret

SPC_WRITE_DSP_ADDR:
    movb %al,_SPC_DSP_ADDR
    ret

SPC_WRITE_DSP_DATA:
    movb %al,_SPC_DSP_DATA
    pusha
    call _SPC_WRITE_DSP
    popa
    ret

SPC_WRITE_PORT0W:
    movb %al,_SPC_PORT0W
    ret
SPC_WRITE_PORT1W:
    movb %al,_SPC_PORT1W
    ret
SPC_WRITE_PORT2W:
    movb %al,_SPC_PORT2W
    ret
SPC_WRITE_PORT3W:
    movb %al,_SPC_PORT3W
    ret

SPC_WRITE_TIMER_0:
    pusha
    SAVE_CYCLES   // Set cycle counter
    call _Update_SPC_Timer_0    // Timer must catch up before changing target
    popa
    testb %al,%al
    movb %al,_SPC_TIMT0 // (0.32) Butcha - timer targets are writable
    setz _SPC_TIMT0+1   // 0 = 256
    ret

SPC_WRITE_TIMER_1:
    pusha
    SAVE_CYCLES   // Set cycle counter
    call _Update_SPC_Timer_1    // Timer must catch up before changing target
    popa
    testb %al,%al
    movb %al,_SPC_TIMT1 // (0.32) Butcha - timer targets are writable
    setz _SPC_TIMT1+1   // 0 = 256
    ret

SPC_WRITE_TIMER_2:
    pusha
    SAVE_CYCLES   // Set cycle counter
    call _Update_SPC_Timer_2    // Timer must catch up before changing target
    popa
    testb %al,%al
    movb %al,_SPC_TIMT2 // (0.32) Butcha - timer targets are writable
    setz _SPC_TIMT2+1   // 0 = 256
    ret

.text; ALIGNC
.data; ALIGND
.bss; ALIGND
