// CPU/interrupt/graphics timing implementation

// Fixed events (always at same time in frame):
// Scanline 0: hidden
// Scanline 1-224/239, dots   0- 21: Hblank
// Scanline 1-224/239, dots  22-277: Displayed screen
// Scanline 1-224/239, dots 278-339: Hblank, HDMA start
// Scanline 225/240: Hblank, HDMA start

// Variable events (software-specified time in frame):
//  H-IRQ - on a specific point on every scanline
//  V-IRQ - short time after beginning of a specific scanline
//  H+V-IRQ - on a specific point on the frame

#include "misc.i"
#include "scankeys.h"
#include "key.i"
#include "CPU.i"
#include "screen.i"
#include "DMA.i"
#include "SPC.i"

#define FPSTicks _FPSTicks
#define FPSMaxTicks _FPSMaxTicks
#define FPSLast _FPSLast

#define FRAME_LIMIT 0
// Set this to force emulation loop to break after x frames have been
//  displayed (not emulated!) This does not save CPU regs, etc.!
//  Primary use is for profiling!
//#define VAR_FRAME_LIMIT
// Set this to force emulation loop to break after FrameLimit frames

.bss
/*
.globl LineBG1OFS,LineBG2OFS,LineBG3OFS,LineBG4OFS,LineM7
ALIGND
LineBG1OFS:     skipw 2*239
ALIGND
LineBG2OFS:     skipw 2*239
ALIGND
LineBG3OFS:     skipw 2*239
ALIGND
LineBG4OFS:     skipw 2*239
ALIGND
LineM7:         skipw 6*239
.globl ScreenLineHoles
ScreenLineHoles:skipb (256+128)/8
*/
.globl LastRenderLine,LastVBLLine
LastRenderLine: skipl
LastVBLLine:    skipl
.globl Ready_Line_Render,ScreenDestPtr
Ready_Line_Render:skipl
ScreenDestPtr:  skipl
.globl FrameCount,Event_Handler,Fixed_Event,Last_Trip,HTimer,HTarget
FrameCount:         skipl   // Used for frameskipping
Event_Handler:      skipl   // Used for render/HDMA timing/NMI/IRQ
Fixed_Event:        skipl   // Used for render/HDMA timing
Last_Trip:          skipl   // Clock cycle position of last scheduled event
HDMA_Next_Event:    skipl   // Fixed event following HDMA
Render_Next_Event:  skipl   // Fixed event following render
Render_Next_Trip:   skipl   // Fixed event following render
Vblank_Start:   skipl       // Scanline of last Vblank start
HTimer:         skipl       // Clock cycle position of H-IRQ delay start
HTarget:        skipl       // Clock cycle position of H-IRQ delay end

.globl HTIMEL,HTIMEH
HTIMEL:     skipb   // H IRQ position low
HTIMEH:     skipb   // H IRQ position high
            skipb
            skipb
.globl VTIMEL,VTIMEH
VTIMEL:     skipb   // V IRQ position low
VTIMEH:     skipb   // V IRQ position high
            skipb
            skipb

.globl FPSTicks,FPSMaxTicks,FPSLast
FPSTicks:       skipl   // Count of timer ticks for this second
FPSCount:       skipl   // Count of frames executed this second
FPSMaxTicks:    skipl   // Number of timer ticks after which FPS is calc'd
FPSLast:        skipl   // Calculated FPS for last second

.text

.globl E1_IRQ,E0_IRQ
E1_IRQ:
    // Emulation mode IRQ
    movl _PC,%eax
    E1_PUSH_W
    CLR_FLAG SNES_FLAG_B    // Clear break bit on stack
    E1_SETUPFLAGS           // put flags into SNES packed flag format
    SET_FLAG SNES_FLAG_B
    E1_PUSH_B

    movl _IRQ_Evector,%eax  // Get Emulation mode IRQ vector
    movl _IRQ_Eoffset,%ebx
    jmp 2f

ALIGNC
E0_IRQ:
    // Native mode IRQ
    movb _PB,%al
    E0_PUSH_B
    movl _PC,%eax
    E0_PUSH_W
    E0_SETUPFLAGS           // put flags into SNES packed flag format
    E0_PUSH_B

    movl _IRQ_Nvector,%eax  // Get Native mode IRQ vector
    movl _IRQ_Noffset,%ebx
2:  movl %eax,_PC           // Setup PC vector
    movl %ebx,_PBOffset
    movb $0,_PB             // Setup bank
    SET_FLAG SNES_FLAG_I    // Disable IRQs
    movl _Read_BankMapping,%eax
    movb $4,CycleTable  // SlowROM bank
    CLR_FLAG SNES_FLAG_D    // Disable decimal mode
    movl %eax,_PB_Read
    ret

.macro Update_Cycles Execute=0
    movl SNES_Cycles,%eax   // Update CPU and SPC cycles
    SAVE_CYCLES
    subl SPC_last_cycles,%eax
    movl _Cycles,SPC_last_cycles
.if \Execute-0
    cmpb $0,_SPC_ENABLED
    jz  7f
    addl SPC_CPU_cycles,%eax
    Execute_SPC SaveCycles=0
7:
.else
    addl %eax,SPC_CPU_cycles
.endif
.endm

.macro Reset_ScreenDestPtr
    movl _SNES_Screen8,%edi // (256+64)*241 framebuffer
    movl %edi,ScreenDestPtr // Setup destination screen pointer
.endm

// Uses 012-local labels, corrupts edx, zeroes eax
.macro Update_FPS_Counter
    movl FPSMaxTicks,%edx
    xorl %eax,%eax
    cmpl %edx,FPSTicks
    jb 0f
    movl FPSCount,%edx
    je 1f
    cmpl $1,%edx
    ja 1f
    movl $0,FPSLast
    jmp 2f
1:  movl %edx,FPSLast
2:  movl %eax,FPSTicks
    movl %eax,FPSCount
ALIGNC
0:
.endm

// CPU/Render/Display/Blanking timing
// Master clock: 21.42MHz   Dot clock: masterclock/4
// Vblank: Lines (DisplayEnd + 1) to 0
// Display: Lines 1 to DisplayEnd
// < 22d Hblank><256d display       >< 62d Hblank> (render)
// <600c exec  >< 40c memory refresh><720c exec  > (execution)

// Keep one fixed event cycle-target/handler pointer
// Keep one variable (may be fixed or IRQ)
.globl IRQNewFrameReset
ALIGNC
IRQNewFrameReset:
    movl $0,Last_Trip

    movl $(278 * 4) - 40,_EventTrip
    movl $(278 * 4) - 40,FixedTrip
    movl $HDMA_Event,Event_Handler
    movl $HDMA_Event,Fixed_Event
    movl $IRQFirstRender,HDMA_Next_Event

    // Reset frameskip counter
    movl $1,%ebx
    cmpl _FRAME_SKIP_MIN,%ebx
    ja  2f
    movl _FRAME_SKIP_MIN,%ebx
2:  movl %ebx,FrameCount

    xorl %eax,%eax
    movl %eax,Throttle      // Reset speed-throttle timer
    movl %eax,Current_Line_Timing   // Reset scanline counter
    incl FPSCount           // For FPS counter - 0.25b14

    RELATCH_HDMA
    Reset_ScreenDestPtr     // Reset framebuffer render address
    Update_FPS_Counter
    ret

ALIGNC
HDMA_Event:
    movl _EventTrip,%eax
    movl %eax,Last_Trip

    movl HDMA_Next_Event,%eax
    movl $(340 * 4) - 40,_EventTrip
    movl $(340 * 4) - 40,FixedTrip
    movl %eax,Event_Handler
    movl %eax,Fixed_Event
    call do_HDMA
    call HIRQ_Check         // chain
    jmp CPU_START           // Return to CPU

ALIGNC
Render_Event:
    movl _EventTrip,%eax
    movl %eax,Last_Trip

    movl Render_Next_Event,%eax
    movl %eax,Event_Handler
    movl %eax,Fixed_Event
    movl Render_Next_Trip,%eax
    movl %eax,_EventTrip
    movl %eax,FixedTrip
    RenderScanline
    call HIRQ_Check         // chain
    jmp CPU_START           // Return to CPU

.globl IRQNewFrame
ALIGNC
2:  hlt
    jmp 1f
ALIGNC
IRQNewFrame:                // Check IRQ, Frame Skip
    movl $0,Last_Trip

    subl %edi,_Cycles
    Update_Cycles Execute=0

    movl $(278 * 4) - 40,_EventTrip
    movl $(278 * 4) - 40,FixedTrip
    movl $HDMA_Event,Event_Handler
    movl $HDMA_Event,Fixed_Event

    movb _INIDISP,%al
    andb $0x0F,%al
    cmpb _BrightnessLevel,%al
    je  0f
    movb $1,PaletteChanged
    movb %al,_BrightnessLevel
0:
    movl _OAMAddress_VBL,%eax
    movl %eax,_OAMAddress   // Restore OAM address every Vblank
    xorl %eax,%eax
    movb %al,OAMOdd         // Reset OAM read/write odd/even selects
    movb %al,OAMReadOdd

    decl FrameCount         // Should we redraw the screen this frame?
    movl %eax,Current_Line_Timing   // Reset scanline counter
    movl $IRQFirst,%edi
    jnz 0f
    incl FPSCount           // For FPS counter - 0.25b14
    movl $IRQFirstRender,%edi
// If minimum frameskip set, don't wait for timer to tell us to draw one
    cmpl %eax,_FRAME_SKIP_MIN
    jnz 2f
// Wait for timer to tell us to draw a frame
1:  movl Throttle,%ebx
    testl %ebx,%ebx
// HLT should be more multi-tasking friendly but appears
// to cause problems with pure DOS in some cases?
    jz  1b  //2b
2:  movl %eax,Throttle      // Reset speed-throttle timer
    cmpl _FRAME_SKIP_MAX,%ebx
    jb  2f
    movl _FRAME_SKIP_MAX,%ebx
2:  cmpl _FRAME_SKIP_MIN,%ebx
    ja  2f
    movl _FRAME_SKIP_MIN,%ebx
2:  movl %ebx,FrameCount    // Reset frame counter
#ifdef DEBUG
    incl _Frames
.if FRAME_LIMIT-0
    cmpl $FRAME_LIMIT,_Frames
    jne 0f
    ret
.endif
#endif
0:  movl %edi,HDMA_Next_Event
    xorb $0x80,STAT78       // Toggle current field
    RELATCH_HDMA
    Reset_ScreenDestPtr     // Reset framebuffer render address
    Update_FPS_Counter
    xorl %eax,%eax
    jmp CPU_START_IRQ       // Return to CPU

ALIGNC
.globl _IRQFirstRender
_IRQFirstRender:
IRQFirstRender: // Check HDMA, IRQ, Render
    movl $0,Last_Trip

    movl $(22 * 4),_EventTrip
    movl $(22 * 4),FixedTrip
    movl $Render_Event,Event_Handler
    movl $Render_Event,Fixed_Event
    movl $HDMA_Event,Render_Next_Event
    movl $(278 * 4) - 40,Render_Next_Trip
    movl $IRQRender,HDMA_Next_Event

    movl $0,Current_Line_Render
    movl $0,Ready_Line_Render

    subl %edi,_Cycles
    Update_Cycles Execute=0
    movb HVBJOY,%al
    andb $0x7F,%al  // VBlank off
    movb %al,HVBJOY
    movb $2,RDNMI           // Clear NMI enabled bit in 0x4210

    incl Current_Line_Timing
    movl Current_Line_Timing,%eax   // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGNC
.globl _IRQRender
_IRQRender:
IRQRender:      // Check HDMA, IRQ, Render
    movl $0,Last_Trip

    subl %edi,_Cycles
    Update_Cycles Execute=0

    incl Current_Line_Timing
    movl Current_Line_Timing,%eax   // Get current scanline
    cmpl %eax,LastRenderLine
    jbe 0f
    movl $(22 * 4),_EventTrip
    movl $(22 * 4),FixedTrip
    movl $Render_Event,Event_Handler
    movl $Render_Event,Fixed_Event
    jmp CPU_START_IRQ       // Return to CPU
ALIGNC
0:  movl $(22 * 4),_EventTrip
    movl $(22 * 4),FixedTrip
    movl $Render_Event,Event_Handler
    movl $Render_Event,Fixed_Event
    movl $NMI,Render_Next_Event
    movl $(340 * 4) - 40,Render_Next_Trip
    jmp CPU_START_IRQ       // Return to CPU

ALIGNC
VBL_Update_Controllers:
    movl $0,Last_Trip

    subl %edi,_Cycles
    Update_Cycles Execute=0

    movl Current_Line_Timing,%eax
    incl %eax
    movl %eax,Current_Line_Timing
    subl Vblank_Start,%eax
    cmpl $3,%eax
    jb  0f

    call UPDATE_CONTROLLERS
    movl $VBL,Event_Handler
    movl $VBL,Fixed_Event
    andb $0xFE,HVBJOY   // Controllers ready
0:  movl Current_Line_Timing,%eax   // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGNC
VBL:movl $0,Last_Trip

    subl %edi,_Cycles
    Update_Cycles Execute=0

    incl Current_Line_Timing
    movl Current_Line_Timing,%eax   // Get current scanline
    cmpl %eax,LastVBLLine   // Needs adjustment for odd line
    ja 0f
    movl $IRQNewFrame,Event_Handler
    movl $IRQNewFrame,Fixed_Event
0:  jmp CPU_START_IRQ       // Return to CPU

ALIGNC
.globl _IRQFirst
_IRQFirst:
IRQFirst:       // Check HDMA, IRQ, Render
    movl $0,Last_Trip

    movl $(278 * 4) - 40,_EventTrip
    movl $(278 * 4) - 40,FixedTrip
    movl $HDMA_Event,Event_Handler
    movl $HDMA_Event,Fixed_Event
    movl $IRQNoRender,HDMA_Next_Event

    movl $1024,Current_Line_Render

    subl %edi,_Cycles
    Update_Cycles Execute=0
    movb HVBJOY,%al
    andb $0x7F,%al  // VBlank off
    movb %al,HVBJOY
    movb $2,RDNMI           // Clear NMI enabled bit in 0x4210

    incl Current_Line_Timing
    movl Current_Line_Timing,%eax   // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

ALIGNC
IRQNoRender:            // Check HDMA, IRQ
    movl $0,Last_Trip

    subl %edi,_Cycles
    Update_Cycles Execute=0

    incl Current_Line_Timing
    movl Current_Line_Timing,%eax   // Get current scanline
    cmpl %eax,LastRenderLine
    jbe 0f
    movl $(278 * 4) - 40,_EventTrip
    movl $(278 * 4) - 40,FixedTrip
    movl $HDMA_Event,Event_Handler
    movl $HDMA_Event,Fixed_Event
    jmp CPU_START_IRQ       // Return to CPU
ALIGNC
0:  movl $NMI_NoRender,Event_Handler
    movl $NMI_NoRender,Fixed_Event
    jmp CPU_START_IRQ       // Return to CPU

ALIGNC
EXPORT_C NMI
EXPORT NMI
    SyncRender  //*
    pushl _Cycles
    movw _Real_SNES_Palette,%ax
    pushl %eax

    movb _fixedpalettecheck,%al
    testb %al,%al
    jz  Palette_Check

Fixed_Back_Color_Add_Hack:
    movw _COLDATA,%ax
    movw %ax,_Real_SNES_Palette
    jmp Do_Palette

Palette_Check:
    cmpb $0,PaletteChanged  // If 1, palette changed!
    jz Do_Copy
Do_Palette:
    movb $0,PaletteChanged

    pusha
    call _SetPalette        // VESA friendly method of setting palette
    popa

Do_Copy:
    cmpb $0,_FPS_ENABLED
    jz  0f
    call _ShowFPS
0:
    call Copy_Screen
//  call _Display_Debug

    popl %eax
    movw %ax,_Real_SNES_Palette
    popl _Cycles
.globl NMI_NoRender,_NMI_NoRender
_NMI_NoRender:
NMI_NoRender:
    movl $0,Last_Trip

    movl $VBL_Update_Controllers,Event_Handler
    movl $VBL_Update_Controllers,Fixed_Event
    subl EventTrip,_Cycles
    // Force SPC to catch up to eliminate lags and improve sound
    Update_Cycles Execute=1
    // Update SPC timers to prevent overflow
    call _Update_SPC_Timer_0
    call _Update_SPC_Timer_1
    call _Update_SPC_Timer_2

//  call UPDATE_CONTROLLERS
#define BUTCHA_SOUND
#ifdef BUTCHA_SOUND
    call _SNDUpdate         // Added by Butcha
#endif

    movb HVBJOY,%al
    orb  $0x81,%al          // VBlank on, controllers being read
    movb %al,HVBJOY
    movb $0x82,%al
    movb %al,RDNMI          // Set NMI enabled bit in 0x4210
    movl Current_Line_Timing,%eax
    incl %eax
    movl %eax,Current_Line_Timing
    movl %eax,Vblank_Start
CheckNMI:
#ifdef DEBUG
    movb _PB,%al
    movb %al,_OLD_PB        // Save program bank
    movl _PC,%eax
    movl %eax,_OLD_PC       // Save program counter
#endif
    testb $0x80,NMITIMEN
    jz  CHECK_KEYS          // NMI off? (via hardware register)

    movl $19,_EventTrip     // Neill's timing doc: 19-23 clock delay
    movl $19,FixedTrip
    movl $NMI_Event,Event_Handler
    movl $NMI_Event,Fixed_Event

CHECK_KEYS:
    xorb %ah,%ah
    JUMP_KEY_UP KEY_F1,0f,%ah   // Disable/enable channel 0
    movb $1,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F1,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F2,0f,%ah   // Disable/enable channel 1
    movb $2,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F2,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F3,0f,%ah   // Disable/enable channel 2
    movb $4,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F3,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F4,0f,%ah   // Disable/enable channel 3
    movb $8,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F4,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F5,0f,%ah   // Disable/enable channel 4
    movb $0x10,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F5,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F6,0f,%ah   // Disable/enable channel 5
    movb $0x20,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F6,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F7,0f,%ah   // Disable/enable channel 6
    movb $0x40,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F7,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_F8,0f,%ah   // Disable/enable channel 7
    movb $0x80,%al
    xorb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_F8,1b,%ah // Wait for key release

0:  JUMP_KEY_UP KEY_1,0f,%ah    // Toggle BG1 disable
    movb $1,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_1,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_2,0f,%ah    // Toggle BG2 disable
    movb $2,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_2,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_3,0f,%ah    // Toggle BG3 disable
    movb $4,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_3,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_4,0f,%ah    // Toggle BG4 disable
    movb $8,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_4,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_5,0f,%ah    // Disable/enable sprites
    movb $0x10,%al
    xorb %al,TM_MASK
1:  JUMP_KEY_DOWN KEY_5,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_7,0f,%ah    // Toggle Offset Change
    call Toggle_Offset_Change
1:  JUMP_KEY_DOWN KEY_7,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_6,0f,%ah    // Reset layers, sprites, channels
    movb $0xFF,%al
    movb $0,Offset_Change_Disable
    movb $0,Layering_Mode
    movb %al,TM_MASK
    movb %al,_SPC_MASK
1:  JUMP_KEY_DOWN KEY_6,1b,%ah  // Wait for key release

0:  JUMP_KEY_UP KEY_8,0f,%ah    // Layering mode
    movb Layering_Mode,%al
    cmpb $2,%al
    incb %al
    jb  2f
    movb $0,%al
2:  movb %al,Layering_Mode
1:  JUMP_KEY_DOWN KEY_8,1b,%ah  // Wait for key release

0:  call Update_Layering    // Update layer mask

    JUMP_KEY_UP KEY_0,0f,%ah    // Save PCX screenshot

    pusha
    call _OutputScreen
    popa

1:  JUMP_KEY_DOWN KEY_0,1b,%ah  // Wait for key release

0:  JUMP_KEY_DOWN KEY_ESC,Emulation_Break,%ah   // Is ESC pressed?

// Return to emulation
    movl Current_Line_Timing,%eax   // Get current scanline
    jmp CPU_START_IRQ       // Return to CPU

Emulation_Break:
0:  JUMP_KEY_DOWN KEY_ESC,0b,%ah    // Wait for key release
    pusha
    movl _PB_Shifted,%ebx
    movw _PC,%bx

    call *_PB_Read
    movb %al,_Map_Byte

    popa

    movl Current_Line_Timing,%eax   // Get current scanline
    CheckVIRQ
    ret

ALIGNC
NMI_Event:
    movl _EventTrip,%eax
    movl %eax,Last_Trip

    movl $(340 * 4) - 40,_EventTrip
    movl $(340 * 4) - 40,FixedTrip
    movl $VBL_Update_Controllers,Event_Handler
    movl $VBL_Update_Controllers,Fixed_Event

    movb In_WAI,%al
    testb %al,%al
    jz  0f
    incw _PC
    movb $0,In_WAI
0:  LOAD_P
    JUMP_NOT_FLAG SNES_FLAG_E,E0_NMI    // Are we in emulation mode?

E1_NMI:
    // Emulation mode NMI
    movl _PC,%eax
    E1_PUSH_W
    CLR_FLAG SNES_FLAG_B    // Clear break bit on stack
    E1_SETUPFLAGS           // put flags into SNES packed flag format
    SET_FLAG SNES_FLAG_B
    E1_PUSH_B

    movl _NMI_Evector,%eax  // Get Emulation mode NMI vector
    movl _NMI_Eoffset,%ebx
    jmp 2f

ALIGNC
E0_NMI:
    // Native mode NMI
    movb _PB,%al
    E0_PUSH_B
    movl _PC,%eax
    E0_PUSH_W
    E0_SETUPFLAGS           // put flags into SNES packed flag format
    E0_PUSH_B

    movl _NMI_Nvector,%eax  // Get Native mode NMI vector
    movl _NMI_Noffset,%ebx
2:  movl %eax,_PC           // Setup PC vector
    movl %ebx,_PBOffset
    movb $0,_PB             // Setup bank
    SET_FLAG SNES_FLAG_I    // Disable IRQs
    movl _Read_BankMapping,%eax
    movb $4,CycleTable  // SlowROM bank
    CLR_FLAG SNES_FLAG_D    // Disable decimal mode
    movl %eax,_PB_Read  // Bank 0 read through memory mapper
    SAVE_P
    call HIRQ_Check         // chain
    jmp CPU_START

ALIGNC
HIRQ_Event:
    movl _EventTrip,%eax
    movl %eax,Last_Trip

    movb $0x40,%dh          // H-IRQ
    testb $0x20,%dl         // Vertical IRQ enabled?
    jz  1f
    movb $0xC0,%dh          // H+V-IRQ
1:  movb $0x80,TIMEUP       // Set IRQ recv'd flag
    orb  %dh,IRQ_pin
    movb In_WAI,%al
    testb %al,%al
    jz  1f
    incw _PC
    movb $0,In_WAI
1:  LOAD_P
    JUMP_FLAG SNES_FLAG_I,0f    // Interrupts disabled?
    pushl $1f
    addl $60,SNES_Cycles    // IRQ processing: 2 IO + 6 bank0
    JUMP_NOT_FLAG SNES_FLAG_E,E0_IRQ
    subl $8,SNES_Cycles     // IRQ processing: 2 IO + 5 bank0
    jmp E1_IRQ
ALIGNC
1:  SAVE_P
0:  movl FixedTrip,%edi
    movl Fixed_Event,%ebx
    movl %edi,_EventTrip
    movl %ebx,Event_Handler
//  call HIRQ_Check         // chain
    jmp CPU_START

ALIGNC
// This can only trash ebx/edi!
HIRQ_Check_Chain:
    movb NMITIMEN,%bl
    testb $0x10,%bl         // H-IRQ enabled?
    jz  0f
    testb $0x20,%bl         // V-IRQ enabled?
    jz  1f
    movl Current_Line_Timing,%edi
    cmpl %edi,VTIMEL
    jne 0f
1:  movl HTimer,%ebx
    movl FixedTrip,%edi
    cmpl %ebx,%edi
    jb  2f
    movl %ebx,_EventTrip
    movl $HIRQ_Event,Event_Handler
    jmp 0f
2:  movl Fixed_Event,%ebx
    movl %edi,_EventTrip
    movl %ebx,Event_Handler
0:  ret

ALIGNC
// This can only trash ebx/edi!
HIRQ_Check:
    movb NMITIMEN,%bl
    testb $0x10,%bl         // H-IRQ enabled?
    jz  0f
    testb $0x20,%bl         // V-IRQ enabled?
    jz  1f
    movl Current_Line_Timing,%edi
    cmpl %edi,VTIMEL
    jne 0f
1:  movl HTimer,%ebx
    movl FixedTrip,%edi
    cmpl %ebx,%edi
    jb  2f
    cmpl %ebx,Last_Trip
    jae 2f
    movl %ebx,_EventTrip
    movl $HIRQ_Event,Event_Handler
    jmp 0f
2:  movl Fixed_Event,%ebx
    movl %edi,_EventTrip
    movl %ebx,Event_Handler
0:  ret

.globl SNES_W4200,SNES_W4207,SNES_W4208,SNES_W4209,SNES_W420A
ALIGNC
SNES_W4200: // NMITIMEN
// a0yx000b  a=NMI on/off,y=vert count,x=horiz count,b=joy read
    cmpb %al,NMITIMEN
    je  0f
    pushl %ebx
    call HIRQ_Check
    movb %al,NMITIMEN
    popl %ebx
0:  andb $0x1F,IRQ_pin
    ret

ALIGNC
SNES_W4207: // HTIMEL
    pushl %ebx
    movl HTIMEL,%ebx
    cmpb %bl,%al
    je  0f
    movb %al,%bl
    movl %ebx,HTIMEL
    shll $2,%ebx

    cmpl $(22+128)*4,%ebx    // Before memory refresh
    jb  1f
    cmpl $(22+128+10)*4,%ebx // During memory refresh
    jb  2f
    subl $10*4,%ebx
    jmp 1f
ALIGNC
2:  movl $(22+128+10)*4,%ebx // During memory refresh
1:  movl %ebx,HTimer
    call HIRQ_Check
0:  popl %ebx
    ret

ALIGNC
SNES_W4208: // HTIMEH
    pushl %ebx
    movl HTIMEL,%ebx
    andb $1,%al
    andb $1,%bh
    cmpb %bh,%al
    je  0f
    movb %al,%bh
    movl %ebx,HTIMEL
    shll $2,%ebx

    cmpl $(22+128)*4,%ebx    // Before memory refresh
    jb  1f
    cmpl $(22+128+10)*4,%ebx // During memory refresh
    jb  2f
    subl $10*4,%ebx
    jmp 1f
ALIGNC
2:  movl $(22+128+10)*4,%ebx // During memory refresh
1:  movl %ebx,HTimer
    call HIRQ_Check
0:  popl %ebx
    ret

ALIGNC
SNES_W4209: // VTIMEL
    pushl %ebx
    movl VTIMEL,%ebx
    cmpb %bl,%al
    je  0f
    movb %al,VTIMEL
    call HIRQ_Check
0:  popl %ebx
    ret

ALIGNC
SNES_W420A: // VTIMEH   // Handled in screen core!
    pushl %ebx
    movl VTIMEL,%ebx
    andb $1,%al
    andb $1,%bh
    cmpb %bh,%al
    je  0f
    movb %al,VTIMEH
    call HIRQ_Check
0:  popl %ebx
    ret

    pushl %eax
    andb $1,%al
    movb %al,VTIMEH
    popl %eax
    ret
