//#define TRACKERS 1048576
/*
 SNEeSe 65c816 CPU emulation core
 Originally written by Savoury SnaX (Not quite sure if I like AT&T)
 Maintained/rewritten by Charles Bilyue'

 Compile under DJGPP (GNU-AS)
  NOTE - This must have .S (capital) or the # stuff wont be pre-processed!!!

 This file contains:
  CPU core info
  Reset
  Execution Loop
  Invalid Opcode Handler
  Flag format conversion tables
  Variable definitions (registers, interrupt vectors, etc.)
  CPU opcode emulation handlers
  CPU opcode handler tables
  CPU opcode timing tables

 CPU core info:
  All general registers are now used in 65c816 emulation:
   EAX,EBX are used by the memory mapper;
   ECX is used to hold P register;
   EDX is used to hold base address to CPU register set;
   EBP is used to hold scanline cycle counter;
   ESI is used by the opcode fetcher;
   EDI is used as memory mapper work register.

    Accumulator         - _A
    X index             - _X
    Y index             - _Y
    Stack pointer       - _S
    Direct address      - _D
    Program Counter     - _PC
    Program Bank        - _PB, _PB_Shifted
    Data Bank           - _DB, _DB_Shifted
    Processor status    - _P,_P_B,_P_W
                True x86 layout = |V|-|-|-|S|Z|-|A|-|-|-|C|
       Native mode 65816 layout =       |E|N|V|M|X|D|I|Z|C|
    Emulation mode 65816 layout =       |E|N|V|1|B|D|I|Z|C|
                            Using       |E|N|Z|I|D|X|M|V|C|
                                                  |B|1|

 CPU timing
  According to Neill Corlett's SNES timing doc, main CPU clock is
   21.47727MHz, 1360 clocks/scanline.

  SNES PPU timing and interrupt timing is specific to ROM country,
   set by ROM loader.
  65c816 runs with many waitstates added in for bus access, bringing
   effective CPU speed between 2.68MHz (minimum) and 3.58MHz (maximum).

  Now adding base opcode cycle counts, *8 for SlowROM banks and *6 for
   FastROM banks. This is INACCURATE! But close enough for now.

 SPC relative timing calculation (for parallel execution)

  SPC runs at 2.048Mhz, for every x cycles of CPU, execute 1 SPC cycle...
   21.48 / 2.048 Mhz = 31.5 CPU cycles per 3 SPC cycles

  SPC execution is on-demand, and CPU and SPC synchronize on every CPU
   event and on CPU<>SPC communication. Executed CPU cycles are counted
   and depleted when SPC execution catches up.

  Executed CPU cycles are divided by 32, multiplied by 3, then added
   to SPC executed cycles target. Upon SPC execution, SPC executes until
   it has reached or passed that target.

   ex: CPU executes 968 cycles:
     x=(968/32)*3; x=30*3; x=90 SPC cycles, added to SPC cycles target

*/

#include "misc.i"
#include "CPU.i"
#include "CPUmem.i"

EXPORT_C CPU_start

.bss
ALIGND
#define R_IRQ_Nvector       -0x40(_Base)
#define R_IRQ_Noffset       -0x3C(_Base)
#define R_NMI_Nvector       -0x38(_Base)
#define R_NMI_Noffset       -0x34(_Base)
#define R_BRK_Nvector       -0x30(_Base)
#define R_BRK_Noffset       -0x2C(_Base)
#define R_COP_Nvector       -0x28(_Base)
#define R_COP_Noffset       -0x24(_Base)
#define R_IRQ_Evector       -0x20(_Base)
#define R_IRQ_Eoffset       -0x1C(_Base)
#define R_NMI_Evector       -0x18(_Base)
#define R_NMI_Eoffset       -0x14(_Base)
#define R_COP_Evector       -0x10(_Base)
#define R_COP_Eoffset       -0x0C(_Base)
#define R_RES_Evector       -0x08(_Base)
#define R_RES_Eoffset       -0x04(_Base)
#define R_PB_Shifted        0x00(_Base)
#define R_PB                0x02(_Base)
#define R_PC                0x04(_Base)
#define R_P                 0x08(_Base)
#define R_SNES_Cycles       0x0C(_Base)
#define R_EventTrip         0x10(_Base)
#define R_A                 0x14(_Base)
#define R_X                 0x18(_Base)
#define R_Y                 0x1C(_Base)
#define R_D                 0x20(_Base)
#define R_S                 0x24(_Base)
#define R_DB_Shifted        0x28(_Base)
#define R_DB                0x2A(_Base)
#define R_OpTable           0x2C(_Base)
#define R_PB_Read           0x30(_Base)
#define R_PBOffset          0x34(_Base)
#define R_LineCycles        0x38(_Base)
#define R_FixedTrip         0x3C(_Base)
#define R_SPC_last_cycles   0x40(_Base)
#define R_SPC_CPU_cycles    0x44(_Base)
#define R_SPC_cycles_left   0x48(_Base)

.globl _NMI_Nvector,_NMI_Noffset,_NMI_Evector,_NMI_Eoffset
.globl _IRQ_Nvector,_IRQ_Noffset,_IRQ_Evector,_IRQ_Eoffset
.globl _COP_Nvector,_COP_Noffset,_COP_Evector,_COP_Eoffset
.globl _BRK_Nvector,_BRK_Noffset,_RES_Evector,_RES_Eoffset
_IRQ_Nvector:skipl
_IRQ_Noffset:skipl
_NMI_Nvector:skipl
_NMI_Noffset:skipl
_BRK_Nvector:skipl
_BRK_Noffset:skipl
_COP_Nvector:skipl
_COP_Noffset:skipl
_IRQ_Evector:skipl
_IRQ_Eoffset:skipl
_NMI_Evector:skipl
_NMI_Eoffset:skipl
_COP_Evector:skipl
_COP_Eoffset:skipl
_RES_Evector:skipl
_RES_Eoffset:skipl

// v0.25 - New system for CPU timings... there are now ten paired tables,
// not five single tables. Each pair of tables has a 256-byte SlowROM
// table first, then a 256-byte FastROM table immediately following.
// The address of the current table pair is OpTable+0x400.

// v0.25 - OpTable holds pointer to current opcode emulation/timing tables,
// removes the need for multiple CPU loops.

CPU_Register_Base:

.globl _PC,_A,_P,_X,_Y,_S,_D,_DB_Shifted,_PB_Shifted,EventTrip
.globl SNES_Cycles,LineCycles,FixedTrip
_PB_Shifted:skipl   // Program Bank
_PC:skipl   // Program Counter
_P: skipl   // Processor status (flags)
SNES_Cycles:skipl   // Scanline cycle count for CPU (0-EventTrip)
EventTrip:  skipl   // Cycle of next event on this scanline
_A: skipl   // Accumulator
_X: skipl   // X and Y indices
_Y: skipl
_D: skipl   // Direct address
_S: skipl   // Stack pointer
_DB_Shifted:skipl   // Data Bank
OpTable:    skipl

// v0.25 - Holds memory mapper read handler for current program bank.
.globl _PB_Read,_PBOffset
_PB_Read:   skipl
_PBOffset:  skipl

LineCycles: skipl   // Cycles per scanline
FixedTrip:  skipl   // Cycle of next fixed event on this scanline

.globl SPC_last_cycles,SPC_CPU_cycles,SPC_cycles_left
SPC_last_cycles:skipl
SPC_CPU_cycles: skipl
SPC_cycles_left:skipl

.globl RealColor0   // For when I make the back color +/- hacks
RealColor0: skipl   // really do +/-

.globl In_WAI
In_WAI: skipb   // nonzero when executing a WAI instruction
.globl IRQ_pin
IRQ_pin:skipb
.globl CPU_Executing
CPU_Executing:  skipb   // nonzero if CPU is executing
.globl TM_MASK
TM_MASK:        skipb   // This is used to force planes to disable!

.globl PaletteChanged
PaletteChanged: skipb   // This is used to decide whether to change palette etc!
.globl Offset_Change_Disable
Offset_Change_Disable:  skipb
.globl _FPS_ENABLED
_FPS_ENABLED:   skipb

.data
#define opcode_clocks(bytes, internal, bank0, bus, speed) \
 ((internal * 6) + (bank0 * 8) + (bus * 6) + \
  ((bytes * 8) - (bytes * speed * 2)))
// bytes, internal operations, bank 0 accesses, other bus accesses, speed
//  speed = 0 for SlowROM, 1 for FastROM
// Direct addressing: +1 IO if DL != 0
// Branch relative: +1 IO if branch taken

ALIGND
OpTableE0:
/* 00 */ .long   E0_BRK       ,E0_ORA_Od_xO,E0_COP    ,E0_ORA_d_s
/* 04 */ .long   E0_TSB_d     ,E0_ORA_d    ,E0_ASL_d  ,E0_ORA_IdI
/* 08 */ .long   E0_PHP       ,E0_ORA_i    ,E0_SLA    ,E0_PHD
/* 0C */ .long   E0_TSB_a     ,E0_ORA_a    ,E0_ASL_a  ,E0_ORA_al
/* 10 */ .long   ALL_BPL      ,E0_ORA_OdO_y,E0_ORA_OdO,E0_ORA_Od_sO_y
/* 14 */ .long   E0_TRB_d     ,E0_ORA_d_x  ,E0_ASL_d_x,E0_ORA_IdI_y
/* 18 */ .long   ALL_CLC      ,E0_ORA_a_y  ,E0_INA    ,E0_TCS
/* 1C */ .long   E0_TRB_a     ,E0_ORA_a_x  ,E0_ASL_a_x,E0_ORA_al_x
/* 20 */ .long   E0_JSR_a     ,E0_AND_Od_xO,E0_JSL_al ,E0_AND_d_s
/* 24 */ .long   E0_BIT_d     ,E0_AND_d    ,E0_ROL_d  ,E0_AND_IdI
/* 28 */ .long   E0_PLP       ,E0_AND_i    ,E0_RLA    ,E0_PLD
/* 2C */ .long   E0_BIT_a     ,E0_AND_a    ,E0_ROL_a  ,E0_AND_al
/* 30 */ .long   ALL_BMI      ,E0_AND_OdO_y,E0_AND_OdO,E0_AND_Od_sO_y
/* 34 */ .long   E0_BIT_d_x   ,E0_AND_d_x  ,E0_ROL_d_x,E0_AND_IdI_y
/* 38 */ .long   ALL_SEC      ,E0_AND_a_y  ,E0_DEA    ,ALL_TSC
/* 3C */ .long   E0_BIT_a_x   ,E0_AND_a_x  ,E0_ROL_a_x,E0_AND_al_x
/* 40 */ .long   E0_RTI       ,E0_EOR_Od_xO,DO_INVALID,E0_EOR_d_s
/* 44 */ .long   E0_MVP       ,E0_EOR_d    ,E0_LSR_d  ,E0_EOR_IdI
/* 48 */ .long   E0_PHA       ,E0_EOR_i    ,E0_SRA    ,E0_PHK
/* 4C */ .long   ALL_JMP_a    ,E0_EOR_a    ,E0_LSR_a  ,E0_EOR_al
/* 50 */ .long   ALL_BVC      ,E0_EOR_OdO_y,E0_EOR_OdO,E0_EOR_Od_sO_y
/* 54 */ .long   E0_MVN       ,E0_EOR_d_x  ,E0_LSR_d_x,E0_EOR_IdI_y
/* 58 */ .long   E0_CLI       ,E0_EOR_a_y  ,E0_PHY    ,ALL_TCD
/* 5C */ .long   ALL_JML_al   ,E0_EOR_a_x  ,E0_LSR_a_x,E0_EOR_al_x
/* 60 */ .long   E0_RTS       ,E0_ADC_Od_xO,E0_PER    ,E0_ADC_d_s
/* 64 */ .long   E0_STZ_d     ,E0_ADC_d    ,E0_ROR_d  ,E0_ADC_IdI
/* 68 */ .long   E0_PLA       ,E0_ADC_i    ,E0_RRA    ,E0_RTL
/* 6C */ .long   ALL_JMP_OaO  ,E0_ADC_a    ,E0_ROR_a  ,E0_ADC_al
/* 70 */ .long   ALL_BVS      ,E0_ADC_OdO_y,E0_ADC_OdO,E0_ADC_Od_sO_y
/* 74 */ .long   E0_STZ_d_x   ,E0_ADC_d_x  ,E0_ROR_d_x,E0_ADC_IdI_y
/* 78 */ .long   ALL_SEI      ,E0_ADC_a_y  ,E0_PLY    ,ALL_TDC
/* 7C */ .long   ALL_JMP_Oa_xO,E0_ADC_a_x  ,E0_ROR_a_x,E0_ADC_al_x
/* 80 */ .long   ALL_BRA      ,E0_STA_Od_xO,ALL_BRL   ,E0_STA_d_s
/* 84 */ .long   E0_STY_d     ,E0_STA_d    ,E0_STX_d  ,E0_STA_IdI
/* 88 */ .long   E0_DEY       ,E0_BIT_i    ,E0_TXA    ,E0_PHB
/* 8C */ .long   E0_STY_a     ,E0_STA_a    ,E0_STX_a  ,E0_STA_al
/* 90 */ .long   ALL_BCC      ,E0_STA_OdO_y,E0_STA_OdO,E0_STA_Od_sO_y
/* 94 */ .long   E0_STY_d_x   ,E0_STA_d_x  ,E0_STX_d_y,E0_STA_IdI_y
/* 98 */ .long   E0_TYA       ,E0_STA_a_y  ,E0_TXS    ,E0_TXY
/* 9C */ .long   E0_STZ_a     ,E0_STA_a_x  ,E0_STZ_a_x,E0_STA_al_x
/* A0 */ .long   E0_LDY_i     ,E0_LDA_Od_xO,E0_LDX_i  ,E0_LDA_d_s
/* A4 */ .long   E0_LDY_d     ,E0_LDA_d    ,E0_LDX_d  ,E0_LDA_IdI
/* A8 */ .long   E0_TAY       ,E0_LDA_i    ,E0_TAX    ,E0_PLB
/* AC */ .long   E0_LDY_a     ,E0_LDA_a    ,E0_LDX_a  ,E0_LDA_al
/* B0 */ .long   ALL_BCS      ,E0_LDA_OdO_y,E0_LDA_OdO,E0_LDA_Od_sO_y
/* B4 */ .long   E0_LDY_d_x   ,E0_LDA_d_x  ,E0_LDX_d_y,E0_LDA_IdI_y
/* B8 */ .long   ALL_CLV      ,E0_LDA_a_y  ,E0_TSX    ,E0_TYX
/* BC */ .long   E0_LDY_a_x   ,E0_LDA_a_x  ,E0_LDX_a_y,E0_LDA_al_x
/* C0 */ .long   E0_CPY_i     ,E0_CMP_Od_xO,E0_REP    ,E0_CMP_d_s
/* C4 */ .long   E0_CPY_d     ,E0_CMP_d    ,E0_DEC_d  ,E0_CMP_IdI
/* C8 */ .long   E0_INY       ,E0_CMP_i    ,E0_DEX    ,ALL_WAI
/* CC */ .long   E0_CPY_a     ,E0_CMP_a    ,E0_DEC_a  ,E0_CMP_al
/* D0 */ .long   ALL_BNE      ,E0_CMP_OdO_y,E0_CMP_OdO,E0_CMP_Od_sO_y
/* D4 */ .long   E0_PEI       ,E0_CMP_d_x  ,E0_DEC_d_x,E0_CMP_IdI_y
/* D8 */ .long   ALL_CLD      ,E0_CMP_a_y  ,E0_PHX    ,DO_INVALID
/* DC */ .long   ALL_JML_OaO  ,E0_CMP_a_x  ,E0_DEC_a_x,E0_CMP_al_x
/* E0 */ .long   E0_CPX_i     ,E0_SBC_Od_xO,E0_SEP    ,E0_SBC_d_s
/* E4 */ .long   E0_CPX_d     ,E0_SBC_d    ,E0_INC_d  ,E0_SBC_IdI
/* E8 */ .long   E0_INX       ,E0_SBC_i    ,ALL_NOP   ,ALL_XBA
/* EC */ .long   E0_CPX_a     ,E0_SBC_a    ,E0_INC_a  ,E0_SBC_al
/* F0 */ .long   ALL_BEQ      ,E0_SBC_OdO_y,E0_SBC_OdO,E0_SBC_Od_sO_y
/* F4 */ .long   E0_PEA       ,E0_SBC_d_x  ,E0_INC_d_x,E0_SBC_IdI_y
/* F8 */ .long   ALL_SED      ,E0_SBC_a_y  ,E0_PLX    ,E0_XCE
/* FC */ .long   E0_JSR_Oa_xO ,E0_SBC_a_x  ,E0_INC_a_x,E0_SBC_al_x

// bytes, internal operations, bank 0 accesses, other bus accesses, speed
//  speed = 0 for SlowROM, 1 for FastROM
CCTableE0:
// SlowROM
/* 00 */ .byte opcode_clocks(2, 0, 6, 0, 0) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 2, 0) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 6, 0, 0) // COP
/* 03 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 4, 0, 0) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 2, 0, 0) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 4, 0, 0) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 2, 0) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHP
/* 09 */ .byte opcode_clocks(3, 0, 0, 0, 0) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 4, 0) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 2, 0) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 4, 0) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 2, 0) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BPL r
/* 11 */ .byte opcode_clocks(2, 1, 2, 2, 0) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 2, 0) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 2, 0) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 4, 0, 0) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 4, 0, 0) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 2, 0) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLC
/* 19 */ .byte opcode_clocks(3, 1, 0, 2, 0) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 0) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 4, 0) // TRB a
/* 1D */ .byte opcode_clocks(3, 1, 0, 2, 0) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 4, 0) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 2, 0) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 0) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 2, 0) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 0) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 2, 0, 0) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 2, 0, 0) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 2, 0, 0) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 4, 0, 0) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 2, 0) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLP
/* 29 */ .byte opcode_clocks(3, 0, 0, 0, 0) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 2, 0) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 2, 0) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 4, 0) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 2, 0) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BMI r
/* 31 */ .byte opcode_clocks(2, 1, 2, 2, 0) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 2, 0) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 2, 0) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 2, 0, 0) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 2, 0, 0) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 4, 0, 0) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 2, 0) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEC
/* 39 */ .byte opcode_clocks(3, 1, 0, 2, 0) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSC
/* 3C */ .byte opcode_clocks(3, 1, 0, 2, 0) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 1, 0, 2, 0) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 4, 0) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 2, 0) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 4, 0, 0) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 2, 0) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 0) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 2, 0, 0) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 2, 0, 0) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 4, 0, 0) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 2, 0) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHA
/* 49 */ .byte opcode_clocks(3, 0, 0, 0, 0) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 0) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 2, 0) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 4, 0) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 2, 0) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVC r
/* 51 */ .byte opcode_clocks(2, 1, 2, 2, 0) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 2, 0) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 2, 0) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 2, 0, 0) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 4, 0, 0) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 2, 0) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLI
/* 59 */ .byte opcode_clocks(3, 1, 0, 2, 0) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 0) // JML al
/* 5D */ .byte opcode_clocks(3, 1, 0, 2, 0) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 4, 0) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 2, 0) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 0) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 2, 0) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 0) // PER
/* 63 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 2, 0, 0) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 2, 0, 0) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 4, 0, 0) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 2, 0) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLA
/* 69 */ .byte opcode_clocks(3, 0, 0, 0, 0) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 0) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 0) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 2, 0) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 4, 0) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 2, 0) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVS r
/* 71 */ .byte opcode_clocks(2, 1, 2, 2, 0) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 2, 0) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 2, 0) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 4, 0, 0) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 2, 0) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEI
/* 79 */ .byte opcode_clocks(3, 1, 0, 2, 0) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 0) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 1, 0, 2, 0) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 4, 0) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 2, 0) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 2, 0) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 0) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 2, 0, 0) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 2, 0, 0) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 2, 0, 0) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 2, 0) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEY
/* 89 */ .byte opcode_clocks(3, 0, 0, 0, 0) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 2, 0) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 2, 0) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 2, 0) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 2, 0) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 2, 0) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 2, 0) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 2, 0) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 2, 0) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 2, 0) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 2, 0) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 2, 0) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 2, 0) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 2, 0) // STA al,x

/* A0 */ .byte opcode_clocks(3, 0, 0, 0, 0) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // LDA (d,x)
/* A2 */ .byte opcode_clocks(3, 0, 0, 0, 0) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 2, 0, 0) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 2, 0, 0) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 2, 0, 0) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAY
/* A9 */ .byte opcode_clocks(3, 0, 0, 0, 0) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 2, 0) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 2, 0) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 2, 0) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 2, 0) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCS r
/* B1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 2, 0) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 2, 0) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLV
/* B9 */ .byte opcode_clocks(3, 1, 0, 2, 0) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYX
/* BC */ .byte opcode_clocks(3, 1, 0, 2, 0) // LDY a,x
/* BD */ .byte opcode_clocks(3, 1, 0, 2, 0) // LDA a,x
/* BE */ .byte opcode_clocks(3, 1, 0, 2, 0) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 2, 0) // LDA al,x

/* C0 */ .byte opcode_clocks(3, 0, 0, 0, 0) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 2, 0, 0) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 2, 0, 0) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 2, 0, 0) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 4, 0, 0) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INY
/* C9 */ .byte opcode_clocks(3, 0, 0, 0, 0) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 0) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 2, 0) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 2, 0) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 4, 0) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 2, 0) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BNE r
/* D1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 2, 0) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 2, 0) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 0) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 2, 0, 0) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 4, 0, 0) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLD
/* D9 */ .byte opcode_clocks(3, 1, 0, 2, 0) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 0) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 0) // JML (a)
/* DD */ .byte opcode_clocks(3, 1, 0, 2, 0) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 4, 0) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 2, 0) // CMP al,x

/* E0 */ .byte opcode_clocks(3, 0, 0, 0, 0) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 2, 0, 0) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 2, 0, 0) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 2, 0, 0) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 4, 0, 0) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INX
/* E9 */ .byte opcode_clocks(3, 0, 0, 0, 0) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 0) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 0) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 2, 0) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 2, 0) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 4, 0) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 2, 0) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BEQ r
/* F1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 2, 0) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 2, 0) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 0) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 2, 0, 0) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 4, 0, 0) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SED
/* F9 */ .byte opcode_clocks(3, 1, 0, 2, 0) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 0) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 0) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 1, 0, 2, 0) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 4, 0) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 2, 0) // SBC al,x

// FastROM
/* 00 */ .byte opcode_clocks(2, 0, 6, 0, 1) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 2, 1) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 6, 0, 1) // COP
/* 03 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 4, 0, 1) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 2, 0, 1) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 4, 0, 1) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 2, 1) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHP
/* 09 */ .byte opcode_clocks(3, 0, 0, 0, 1) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 4, 1) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 2, 1) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 4, 1) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 2, 1) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BPL r
/* 11 */ .byte opcode_clocks(2, 1, 2, 2, 1) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 2, 1) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 2, 1) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 4, 0, 1) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 4, 0, 1) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 2, 1) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLC
/* 19 */ .byte opcode_clocks(3, 1, 0, 2, 1) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 1) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 4, 1) // TRB a
/* 1D */ .byte opcode_clocks(3, 1, 0, 2, 1) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 4, 1) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 2, 1) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 1) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 2, 1) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 1) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 2, 0, 1) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 2, 0, 1) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 2, 0, 1) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 4, 0, 1) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 2, 1) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLP
/* 29 */ .byte opcode_clocks(3, 0, 0, 0, 1) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 2, 1) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 2, 1) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 4, 1) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 2, 1) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BMI r
/* 31 */ .byte opcode_clocks(2, 1, 2, 2, 1) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 2, 1) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 2, 1) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 2, 0, 1) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 2, 0, 1) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 4, 0, 1) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 2, 1) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEC
/* 39 */ .byte opcode_clocks(3, 1, 0, 2, 1) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSC
/* 3C */ .byte opcode_clocks(3, 1, 0, 2, 1) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 1, 0, 2, 1) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 4, 1) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 2, 1) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 4, 0, 1) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 2, 1) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 1) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 2, 0, 1) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 2, 0, 1) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 4, 0, 1) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 2, 1) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHA
/* 49 */ .byte opcode_clocks(3, 0, 0, 0, 1) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 1) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 2, 1) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 4, 1) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 2, 1) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVC r
/* 51 */ .byte opcode_clocks(2, 1, 2, 2, 1) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 2, 1) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 2, 1) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 2, 0, 1) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 4, 0, 1) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 2, 1) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLI
/* 59 */ .byte opcode_clocks(3, 1, 0, 2, 1) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 1) // JML al
/* 5D */ .byte opcode_clocks(3, 1, 0, 2, 1) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 4, 1) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 2, 1) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 1) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 2, 1) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 1) // PER
/* 63 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 2, 0, 1) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 2, 0, 1) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 4, 0, 1) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 2, 1) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLA
/* 69 */ .byte opcode_clocks(3, 0, 0, 0, 1) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 1) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 1) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 2, 1) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 4, 1) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 2, 1) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVS r
/* 71 */ .byte opcode_clocks(2, 1, 2, 2, 1) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 2, 1) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 2, 1) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 4, 0, 1) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 2, 1) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEI
/* 79 */ .byte opcode_clocks(3, 1, 0, 2, 1) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 1) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 1, 0, 2, 1) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 4, 1) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 2, 1) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 2, 1) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 1) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 2, 0, 1) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 2, 0, 1) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 2, 0, 1) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 2, 1) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEY
/* 89 */ .byte opcode_clocks(3, 0, 0, 0, 1) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 2, 1) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 2, 1) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 2, 1) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 2, 1) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 2, 1) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 2, 1) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 2, 1) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 2, 1) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 2, 1) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 2, 1) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 2, 1) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 2, 1) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 2, 1) // STA al,x

/* A0 */ .byte opcode_clocks(3, 0, 0, 0, 1) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // LDA (d,x)
/* A2 */ .byte opcode_clocks(3, 0, 0, 0, 1) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 2, 0, 1) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 2, 0, 1) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 2, 0, 1) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAY
/* A9 */ .byte opcode_clocks(3, 0, 0, 0, 1) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 2, 1) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 2, 1) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 2, 1) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 2, 1) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCS r
/* B1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 2, 1) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 2, 1) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLV
/* B9 */ .byte opcode_clocks(3, 1, 0, 2, 1) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYX
/* BC */ .byte opcode_clocks(3, 1, 0, 2, 1) // LDY a,x
/* BD */ .byte opcode_clocks(3, 1, 0, 2, 1) // LDA a,x
/* BE */ .byte opcode_clocks(3, 1, 0, 2, 1) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 2, 1) // LDA al,x

/* C0 */ .byte opcode_clocks(3, 0, 0, 0, 1) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 2, 0, 1) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 2, 0, 1) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 2, 0, 1) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 4, 0, 1) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INY
/* C9 */ .byte opcode_clocks(3, 0, 0, 0, 1) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 1) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 2, 1) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 2, 1) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 4, 1) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 2, 1) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BNE r
/* D1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 2, 1) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 2, 1) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 1) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 2, 0, 1) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 4, 0, 1) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLD
/* D9 */ .byte opcode_clocks(3, 1, 0, 2, 1) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 1) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 1) // JML (a)
/* DD */ .byte opcode_clocks(3, 1, 0, 2, 1) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 4, 1) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 2, 1) // CMP al,x

/* E0 */ .byte opcode_clocks(3, 0, 0, 0, 1) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 2, 0, 1) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 2, 0, 1) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 2, 0, 1) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 4, 0, 1) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INX
/* E9 */ .byte opcode_clocks(3, 0, 0, 0, 1) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 1) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 1) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 2, 1) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 2, 1) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 4, 1) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 2, 1) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BEQ r
/* F1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 2, 1) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 2, 1) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 1) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 2, 0, 1) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 4, 0, 1) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SED
/* F9 */ .byte opcode_clocks(3, 1, 0, 2, 1) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 1) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 1) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 1, 0, 2, 1) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 4, 1) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 2, 1) // SBC al,x

OpTablePm:
/* 00 */ .long   E0_BRK       ,EM_ORA_Od_xO,E0_COP    ,EM_ORA_d_s
/* 04 */ .long   EM_TSB_d     ,EM_ORA_d    ,EM_ASL_d  ,EM_ORA_IdI
/* 08 */ .long   E0_PHP       ,EM_ORA_i    ,EM_SLA    ,E0_PHD
/* 0C */ .long   EM_TSB_a     ,EM_ORA_a    ,EM_ASL_a  ,EM_ORA_al
/* 10 */ .long   ALL_BPL      ,EM_ORA_OdO_y,EM_ORA_OdO,EM_ORA_Od_sO_y
/* 14 */ .long   EM_TRB_d     ,EM_ORA_d_x  ,EM_ASL_d_x,EM_ORA_IdI_y
/* 18 */ .long   ALL_CLC      ,EM_ORA_a_y  ,EM_INA    ,E0_TCS
/* 1C */ .long   EM_TRB_a     ,EM_ORA_a_x  ,EM_ASL_a_x,EM_ORA_al_x
/* 20 */ .long   E0_JSR_a     ,EM_AND_Od_xO,E0_JSL_al ,EM_AND_d_s
/* 24 */ .long   EM_BIT_d     ,EM_AND_d    ,EM_ROL_d  ,EM_AND_IdI
/* 28 */ .long   E0_PLP       ,EM_AND_i    ,EM_RLA    ,E0_PLD
/* 2C */ .long   EM_BIT_a     ,EM_AND_a    ,EM_ROL_a  ,EM_AND_al
/* 30 */ .long   ALL_BMI      ,EM_AND_OdO_y,EM_AND_OdO,EM_AND_Od_sO_y
/* 34 */ .long   EM_BIT_d_x   ,EM_AND_d_x  ,EM_ROL_d_x,EM_AND_IdI_y
/* 38 */ .long   ALL_SEC      ,EM_AND_a_y  ,EM_DEA    ,ALL_TSC
/* 3C */ .long   EM_BIT_a_x   ,EM_AND_a_x  ,EM_ROL_a_x,EM_AND_al_x
/* 40 */ .long   E0_RTI       ,EM_EOR_Od_xO,DO_INVALID,EM_EOR_d_s
/* 44 */ .long   E0_MVP       ,EM_EOR_d    ,EM_LSR_d  ,EM_EOR_IdI
/* 48 */ .long   EM_PHA       ,EM_EOR_i    ,EM_SRA    ,E0_PHK
/* 4C */ .long   ALL_JMP_a    ,EM_EOR_a    ,EM_LSR_a  ,EM_EOR_al
/* 50 */ .long   ALL_BVC      ,EM_EOR_OdO_y,EM_EOR_OdO,EM_EOR_Od_sO_y
/* 54 */ .long   E0_MVN       ,EM_EOR_d_x  ,EM_LSR_d_x,EM_EOR_IdI_y
/* 58 */ .long   E0_CLI       ,EM_EOR_a_y  ,E0_PHY    ,ALL_TCD
/* 5C */ .long   ALL_JML_al   ,EM_EOR_a_x  ,EM_LSR_a_x,EM_EOR_al_x
/* 60 */ .long   E0_RTS       ,EM_ADC_Od_xO,E0_PER    ,EM_ADC_d_s
/* 64 */ .long   EM_STZ_d     ,EM_ADC_d    ,EM_ROR_d  ,EM_ADC_IdI
/* 68 */ .long   EM_PLA       ,EM_ADC_i    ,EM_RRA    ,E0_RTL
/* 6C */ .long   ALL_JMP_OaO  ,EM_ADC_a    ,EM_ROR_a  ,EM_ADC_al
/* 70 */ .long   ALL_BVS      ,EM_ADC_OdO_y,EM_ADC_OdO,EM_ADC_Od_sO_y
/* 74 */ .long   EM_STZ_d_x   ,EM_ADC_d_x  ,EM_ROR_d_x,EM_ADC_IdI_y
/* 78 */ .long   ALL_SEI      ,EM_ADC_a_y  ,E0_PLY    ,ALL_TDC
/* 7C */ .long   ALL_JMP_Oa_xO,EM_ADC_a_x  ,EM_ROR_a_x,EM_ADC_al_x
/* 80 */ .long   ALL_BRA      ,EM_STA_Od_xO,ALL_BRL   ,EM_STA_d_s
/* 84 */ .long   E0_STY_d     ,EM_STA_d    ,E0_STX_d  ,EM_STA_IdI
/* 88 */ .long   E0_DEY       ,EM_BIT_i    ,EM_TXA    ,E0_PHB
/* 8C */ .long   E0_STY_a     ,EM_STA_a    ,E0_STX_a  ,EM_STA_al
/* 90 */ .long   ALL_BCC      ,EM_STA_OdO_y,EM_STA_OdO,EM_STA_Od_sO_y
/* 94 */ .long   E0_STY_d_x   ,EM_STA_d_x  ,E0_STX_d_y,EM_STA_IdI_y
/* 98 */ .long   EM_TYA       ,EM_STA_a_y  ,E0_TXS    ,E0_TXY
/* 9C */ .long   EM_STZ_a     ,EM_STA_a_x  ,EM_STZ_a_x,EM_STA_al_x
/* A0 */ .long   E0_LDY_i     ,EM_LDA_Od_xO,E0_LDX_i  ,EM_LDA_d_s
/* A4 */ .long   E0_LDY_d     ,EM_LDA_d    ,E0_LDX_d  ,EM_LDA_IdI
/* A8 */ .long   E0_TAY       ,EM_LDA_i    ,E0_TAX    ,E0_PLB
/* AC */ .long   E0_LDY_a     ,EM_LDA_a    ,E0_LDX_a  ,EM_LDA_al
/* B0 */ .long   ALL_BCS      ,EM_LDA_OdO_y,EM_LDA_OdO,EM_LDA_Od_sO_y
/* B4 */ .long   E0_LDY_d_x   ,EM_LDA_d_x  ,E0_LDX_d_y,EM_LDA_IdI_y
/* B8 */ .long   ALL_CLV      ,EM_LDA_a_y  ,E0_TSX    ,E0_TYX
/* BC */ .long   E0_LDY_a_x   ,EM_LDA_a_x  ,E0_LDX_a_y,EM_LDA_al_x
/* C0 */ .long   E0_CPY_i     ,EM_CMP_Od_xO,E0_REP    ,EM_CMP_d_s
/* C4 */ .long   E0_CPY_d     ,EM_CMP_d    ,EM_DEC_d  ,EM_CMP_IdI
/* C8 */ .long   E0_INY       ,EM_CMP_i    ,E0_DEX    ,ALL_WAI
/* CC */ .long   E0_CPY_a     ,EM_CMP_a    ,EM_DEC_a  ,EM_CMP_al
/* D0 */ .long   ALL_BNE      ,EM_CMP_OdO_y,EM_CMP_OdO,EM_CMP_Od_sO_y
/* D4 */ .long   E0_PEI       ,EM_CMP_d_x  ,EM_DEC_d_x,EM_CMP_IdI_y
/* D8 */ .long   ALL_CLD      ,EM_CMP_a_y  ,E0_PHX    ,DO_INVALID
/* DC */ .long   ALL_JML_OaO  ,EM_CMP_a_x  ,EM_DEC_a_x,EM_CMP_al_x
/* E0 */ .long   E0_CPX_i     ,EM_SBC_Od_xO,E0_SEP    ,EM_SBC_d_s
/* E4 */ .long   E0_CPX_d     ,EM_SBC_d    ,EM_INC_d  ,EM_SBC_IdI
/* E8 */ .long   E0_INX       ,EM_SBC_i    ,ALL_NOP   ,ALL_XBA
/* EC */ .long   E0_CPX_a     ,EM_SBC_a    ,EM_INC_a  ,EM_SBC_al
/* F0 */ .long   ALL_BEQ      ,EM_SBC_OdO_y,EM_SBC_OdO,EM_SBC_Od_sO_y
/* F4 */ .long   E0_PEA       ,EM_SBC_d_x  ,EM_INC_d_x,EM_SBC_IdI_y
/* F8 */ .long   ALL_SED      ,EM_SBC_a_y  ,E0_PLX    ,E0_XCE
/* FC */ .long   E0_JSR_Oa_xO ,EM_SBC_a_x  ,EM_INC_a_x,EM_SBC_al_x

// bytes, internal operations, bank 0 accesses, other bus accesses, speed
//  speed = 0 for SlowROM, 1 for FastROM
CCTablePm:
// SlowROM
/* 00 */ .byte opcode_clocks(2, 0, 6, 0, 0) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 1, 0) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 6, 0, 0) // COP
/* 03 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 2, 0, 0) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 1, 0, 0) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHP
/* 09 */ .byte opcode_clocks(2, 0, 0, 0, 0) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 2, 0) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BPL r
/* 11 */ .byte opcode_clocks(2, 1, 2, 1, 0) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 1, 0) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 2, 0, 0) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLC
/* 19 */ .byte opcode_clocks(3, 1, 0, 1, 0) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 0) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 2, 0) // TRB a
/* 1D */ .byte opcode_clocks(3, 1, 0, 1, 0) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 0) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 1, 0) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 0) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 1, 0, 0) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 1, 0, 0) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 1, 0, 0) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 1, 0) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLP
/* 29 */ .byte opcode_clocks(2, 0, 0, 0, 0) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 1, 0) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 1, 0) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 1, 0) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BMI r
/* 31 */ .byte opcode_clocks(2, 1, 2, 1, 0) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 1, 0) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 1, 0) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 1, 0, 0) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 1, 0, 0) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 1, 0) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEC
/* 39 */ .byte opcode_clocks(3, 1, 0, 1, 0) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSC
/* 3C */ .byte opcode_clocks(3, 1, 0, 1, 0) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 1, 0, 1, 0) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 1, 0) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 4, 0, 0) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 1, 0) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 0) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 1, 0, 0) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 1, 0, 0) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 1, 0) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHA
/* 49 */ .byte opcode_clocks(2, 0, 0, 0, 0) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 0) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 1, 0) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 2, 0) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 1, 0) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVC r
/* 51 */ .byte opcode_clocks(2, 1, 2, 1, 0) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 1, 0) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 1, 0) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 1, 0, 0) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 2, 0, 0) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 1, 0) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLI
/* 59 */ .byte opcode_clocks(3, 1, 0, 1, 0) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 0) // JML al
/* 5D */ .byte opcode_clocks(3, 1, 0, 1, 0) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 2, 0) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 1, 0) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 0) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 1, 0) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 0) // PER
/* 63 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 1, 0, 0) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLA
/* 69 */ .byte opcode_clocks(2, 0, 0, 0, 0) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 0) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 0) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVS r
/* 71 */ .byte opcode_clocks(2, 1, 2, 1, 0) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 1, 0) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEI
/* 79 */ .byte opcode_clocks(3, 1, 0, 1, 0) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 0) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 1, 0, 1, 0) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 1, 0) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 0) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 2, 0, 0) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 2, 0, 0) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 1, 0) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEY
/* 89 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 2, 0) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 1, 0) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 2, 0) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 1, 0) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 1, 0) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 1, 0) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 1, 0) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 1, 0) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 1, 0) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 1, 0) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 1, 0) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 1, 0) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 1, 0) // STA al,x

/* A0 */ .byte opcode_clocks(3, 0, 0, 0, 0) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // LDA (d,x)
/* A2 */ .byte opcode_clocks(3, 0, 0, 0, 0) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 2, 0, 0) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 2, 0, 0) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAY
/* A9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 2, 0) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 2, 0) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 1, 0) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCS r
/* B1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLV
/* B9 */ .byte opcode_clocks(3, 1, 0, 1, 0) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYX
/* BC */ .byte opcode_clocks(3, 1, 0, 2, 0) // LDY a,x
/* BD */ .byte opcode_clocks(3, 1, 0, 1, 0) // LDA a,x
/* BE */ .byte opcode_clocks(3, 1, 0, 2, 0) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 1, 0) // LDA al,x

/* C0 */ .byte opcode_clocks(3, 0, 0, 0, 0) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 2, 0, 0) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 2, 0, 0) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INY
/* C9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 0) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 2, 0) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 1, 0) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 2, 0) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 1, 0) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BNE r
/* D1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 0) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 2, 0, 0) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLD
/* D9 */ .byte opcode_clocks(3, 1, 0, 1, 0) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 0) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 0) // JML (a)
/* DD */ .byte opcode_clocks(3, 1, 0, 1, 0) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 2, 0) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 1, 0) // CMP al,x

/* E0 */ .byte opcode_clocks(3, 0, 0, 0, 0) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 2, 0, 0) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 2, 0, 0) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INX
/* E9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 0) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 0) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 2, 0) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 1, 0) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 2, 0) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 1, 0) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BEQ r
/* F1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 0) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 2, 0, 0) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SED
/* F9 */ .byte opcode_clocks(3, 1, 0, 1, 0) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 0) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 0) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 1, 0, 1, 0) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 2, 0) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 1, 0) // SBC al,x

// FastROM
/* 00 */ .byte opcode_clocks(2, 0, 6, 0, 1) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 1, 1) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 6, 0, 1) // COP
/* 03 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 2, 0, 1) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 1, 0, 1) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHP
/* 09 */ .byte opcode_clocks(2, 0, 0, 0, 1) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 2, 1) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BPL r
/* 11 */ .byte opcode_clocks(2, 1, 2, 1, 1) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 1, 1) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 2, 0, 1) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLC
/* 19 */ .byte opcode_clocks(3, 1, 0, 1, 1) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 1) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 2, 1) // TRB a
/* 1D */ .byte opcode_clocks(3, 1, 0, 1, 1) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 1) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 1, 1) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 1) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 1, 0, 1) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 1, 0, 1) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 1, 0, 1) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 1, 1) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLP
/* 29 */ .byte opcode_clocks(2, 0, 0, 0, 1) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 1, 1) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 1, 1) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 1, 1) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BMI r
/* 31 */ .byte opcode_clocks(2, 1, 2, 1, 1) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 1, 1) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 1, 1) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 1, 0, 1) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 1, 0, 1) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 1, 1) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEC
/* 39 */ .byte opcode_clocks(3, 1, 0, 1, 1) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSC
/* 3C */ .byte opcode_clocks(3, 1, 0, 1, 1) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 1, 0, 1, 1) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 1, 1) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 4, 0, 1) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 1, 1) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 1) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 1, 0, 1) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 1, 0, 1) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 1, 1) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHA
/* 49 */ .byte opcode_clocks(2, 0, 0, 0, 1) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 1) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 1, 1) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 2, 1) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 1, 1) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVC r
/* 51 */ .byte opcode_clocks(2, 1, 2, 1, 1) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 1, 1) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 1, 1) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 1, 0, 1) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 2, 0, 1) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 1, 1) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLI
/* 59 */ .byte opcode_clocks(3, 1, 0, 1, 1) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 1) // JML al
/* 5D */ .byte opcode_clocks(3, 1, 0, 1, 1) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 2, 1) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 1, 1) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 1) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 1, 1) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 1) // PER
/* 63 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 1, 0, 1) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLA
/* 69 */ .byte opcode_clocks(2, 0, 0, 0, 1) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 1) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 1) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVS r
/* 71 */ .byte opcode_clocks(2, 1, 2, 1, 1) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 1, 1) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEI
/* 79 */ .byte opcode_clocks(3, 1, 0, 1, 1) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 1) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 1, 0, 1, 1) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 1, 1) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 1) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 2, 0, 1) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 2, 0, 1) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 1, 1) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEY
/* 89 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 2, 1) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 1, 1) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 2, 1) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 1, 1) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 1, 1) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 1, 1) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 1, 1) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 1, 1) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 1, 1) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 1, 1) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 1, 1) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 1, 1) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 1, 1) // STA al,x

/* A0 */ .byte opcode_clocks(3, 0, 0, 0, 1) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // LDA (d,x)
/* A2 */ .byte opcode_clocks(3, 0, 0, 0, 1) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 2, 0, 1) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 2, 0, 1) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAY
/* A9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 2, 1) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 2, 1) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 1, 1) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCS r
/* B1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLV
/* B9 */ .byte opcode_clocks(3, 1, 0, 1, 1) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYX
/* BC */ .byte opcode_clocks(3, 1, 0, 2, 1) // LDY a,x
/* BD */ .byte opcode_clocks(3, 1, 0, 1, 1) // LDA a,x
/* BE */ .byte opcode_clocks(3, 1, 0, 2, 1) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 1, 1) // LDA al,x

/* C0 */ .byte opcode_clocks(3, 0, 0, 0, 1) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 2, 0, 1) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 2, 0, 1) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INY
/* C9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 1) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 2, 1) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 1, 1) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 2, 1) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 1, 1) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BNE r
/* D1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 1) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 2, 0, 1) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLD
/* D9 */ .byte opcode_clocks(3, 1, 0, 1, 1) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 1) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 1) // JML (a)
/* DD */ .byte opcode_clocks(3, 1, 0, 1, 1) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 2, 1) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 1, 1) // CMP al,x

/* E0 */ .byte opcode_clocks(3, 0, 0, 0, 1) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 2, 0, 1) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 2, 0, 1) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INX
/* E9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 1) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 1) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 2, 1) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 1, 1) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 2, 1) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 1, 1) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BEQ r
/* F1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 1) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 2, 0, 1) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SED
/* F9 */ .byte opcode_clocks(3, 1, 0, 1, 1) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 1) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 1) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 1, 0, 1, 1) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 2, 1) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 1, 1) // SBC al,x

OpTablePx:
/* 00 */ .long   E0_BRK       ,E0_ORA_Od_xO,E0_COP    ,E0_ORA_d_s
/* 04 */ .long   E0_TSB_d     ,E0_ORA_d    ,E0_ASL_d  ,E0_ORA_IdI
/* 08 */ .long   E0_PHP       ,E0_ORA_i    ,E0_SLA    ,E0_PHD
/* 0C */ .long   E0_TSB_a     ,E0_ORA_a    ,E0_ASL_a  ,E0_ORA_al
/* 10 */ .long   ALL_BPL      ,E0_ORA_OdO_y,E0_ORA_OdO,E0_ORA_Od_sO_y
/* 14 */ .long   E0_TRB_d     ,E0_ORA_d_x  ,E0_ASL_d_x,E0_ORA_IdI_y
/* 18 */ .long   ALL_CLC      ,E0_ORA_a_y  ,E0_INA    ,E0_TCS
/* 1C */ .long   E0_TRB_a     ,E0_ORA_a_x  ,E0_ASL_a_x,E0_ORA_al_x
/* 20 */ .long   E0_JSR_a     ,E0_AND_Od_xO,E0_JSL_al ,E0_AND_d_s
/* 24 */ .long   E0_BIT_d     ,E0_AND_d    ,E0_ROL_d  ,E0_AND_IdI
/* 28 */ .long   E0_PLP       ,E0_AND_i    ,E0_RLA    ,E0_PLD
/* 2C */ .long   E0_BIT_a     ,E0_AND_a    ,E0_ROL_a  ,E0_AND_al
/* 30 */ .long   ALL_BMI      ,E0_AND_OdO_y,E0_AND_OdO,E0_AND_Od_sO_y
/* 34 */ .long   E0_BIT_d_x   ,E0_AND_d_x  ,E0_ROL_d_x,E0_AND_IdI_y
/* 38 */ .long   ALL_SEC      ,E0_AND_a_y  ,E0_DEA    ,ALL_TSC
/* 3C */ .long   E0_BIT_a_x   ,E0_AND_a_x  ,E0_ROL_a_x,E0_AND_al_x
/* 40 */ .long   E0_RTI       ,E0_EOR_Od_xO,DO_INVALID,E0_EOR_d_s
/* 44 */ .long   EX_MVP       ,E0_EOR_d    ,E0_LSR_d  ,E0_EOR_IdI
/* 48 */ .long   E0_PHA       ,E0_EOR_i    ,E0_SRA    ,E0_PHK
/* 4C */ .long   ALL_JMP_a    ,E0_EOR_a    ,E0_LSR_a  ,E0_EOR_al
/* 50 */ .long   ALL_BVC      ,E0_EOR_OdO_y,E0_EOR_OdO,E0_EOR_Od_sO_y
/* 54 */ .long   EX_MVN       ,E0_EOR_d_x  ,E0_LSR_d_x,E0_EOR_IdI_y
/* 58 */ .long   E0_CLI       ,E0_EOR_a_y  ,EX_PHY    ,ALL_TCD
/* 5C */ .long   ALL_JML_al   ,E0_EOR_a_x  ,E0_LSR_a_x,E0_EOR_al_x
/* 60 */ .long   E0_RTS       ,E0_ADC_Od_xO,E0_PER    ,E0_ADC_d_s
/* 64 */ .long   E0_STZ_d     ,E0_ADC_d    ,E0_ROR_d  ,E0_ADC_IdI
/* 68 */ .long   E0_PLA       ,E0_ADC_i    ,E0_RRA    ,E0_RTL
/* 6C */ .long   ALL_JMP_OaO  ,E0_ADC_a    ,E0_ROR_a  ,E0_ADC_al
/* 70 */ .long   ALL_BVS      ,E0_ADC_OdO_y,E0_ADC_OdO,E0_ADC_Od_sO_y
/* 74 */ .long   E0_STZ_d_x   ,E0_ADC_d_x  ,E0_ROR_d_x,E0_ADC_IdI_y
/* 78 */ .long   ALL_SEI      ,E0_ADC_a_y  ,EX_PLY    ,ALL_TDC
/* 7C */ .long   ALL_JMP_Oa_xO,E0_ADC_a_x  ,E0_ROR_a_x,E0_ADC_al_x
/* 80 */ .long   ALL_BRA      ,E0_STA_Od_xO,ALL_BRL   ,E0_STA_d_s
/* 84 */ .long   EX_STY_d     ,E0_STA_d    ,EX_STX_d  ,E0_STA_IdI
/* 88 */ .long   EX_DEY       ,E0_BIT_i    ,E0_TXA    ,E0_PHB
/* 8C */ .long   EX_STY_a     ,E0_STA_a    ,EX_STX_a  ,E0_STA_al
/* 90 */ .long   ALL_BCC      ,E0_STA_OdO_y,E0_STA_OdO,E0_STA_Od_sO_y
/* 94 */ .long   EX_STY_d_x   ,E0_STA_d_x  ,EX_STX_d_y,E0_STA_IdI_y
/* 98 */ .long   E0_TYA       ,E0_STA_a_y  ,E0_TXS    ,EX_TXY
/* 9C */ .long   E0_STZ_a     ,E0_STA_a_x  ,E0_STZ_a_x,E0_STA_al_x
/* A0 */ .long   EX_LDY_i     ,E0_LDA_Od_xO,EX_LDX_i  ,E0_LDA_d_s
/* A4 */ .long   EX_LDY_d     ,E0_LDA_d    ,EX_LDX_d  ,E0_LDA_IdI
/* A8 */ .long   EX_TAY       ,E0_LDA_i    ,EX_TAX    ,E0_PLB
/* AC */ .long   EX_LDY_a     ,E0_LDA_a    ,EX_LDX_a  ,E0_LDA_al
/* B0 */ .long   ALL_BCS      ,E0_LDA_OdO_y,E0_LDA_OdO,E0_LDA_Od_sO_y
/* B4 */ .long   EX_LDY_d_x   ,E0_LDA_d_x  ,EX_LDX_d_y,E0_LDA_IdI_y
/* B8 */ .long   ALL_CLV      ,E0_LDA_a_y  ,EX_TSX    ,EX_TYX
/* BC */ .long   EX_LDY_a_x   ,E0_LDA_a_x  ,EX_LDX_a_y,E0_LDA_al_x
/* C0 */ .long   EX_CPY_i     ,E0_CMP_Od_xO,E0_REP    ,E0_CMP_d_s
/* C4 */ .long   EX_CPY_d     ,E0_CMP_d    ,E0_DEC_d  ,E0_CMP_IdI
/* C8 */ .long   EX_INY       ,E0_CMP_i    ,EX_DEX    ,ALL_WAI
/* CC */ .long   EX_CPY_a     ,E0_CMP_a    ,E0_DEC_a  ,E0_CMP_al
/* D0 */ .long   ALL_BNE      ,E0_CMP_OdO_y,E0_CMP_OdO,E0_CMP_Od_sO_y
/* D4 */ .long   E0_PEI       ,E0_CMP_d_x  ,E0_DEC_d_x,E0_CMP_IdI_y
/* D8 */ .long   ALL_CLD      ,E0_CMP_a_y  ,EX_PHX    ,DO_INVALID
/* DC */ .long   ALL_JML_OaO  ,E0_CMP_a_x  ,E0_DEC_a_x,E0_CMP_al_x
/* E0 */ .long   EX_CPX_i     ,E0_SBC_Od_xO,E0_SEP    ,E0_SBC_d_s
/* E4 */ .long   EX_CPX_d     ,E0_SBC_d    ,E0_INC_d  ,E0_SBC_IdI
/* E8 */ .long   EX_INX       ,E0_SBC_i    ,ALL_NOP   ,ALL_XBA
/* EC */ .long   EX_CPX_a     ,E0_SBC_a    ,E0_INC_a  ,E0_SBC_al
/* F0 */ .long   ALL_BEQ      ,E0_SBC_OdO_y,E0_SBC_OdO,E0_SBC_Od_sO_y
/* F4 */ .long   E0_PEA       ,E0_SBC_d_x  ,E0_INC_d_x,E0_SBC_IdI_y
/* F8 */ .long   ALL_SED      ,E0_SBC_a_y  ,EX_PLX    ,E0_XCE
/* FC */ .long   E0_JSR_Oa_xO ,E0_SBC_a_x  ,E0_INC_a_x,E0_SBC_al_x

// bytes, internal operations, bank 0 accesses, other bus accesses, speed
//  speed = 0 for SlowROM, 1 for FastROM
CCTablePx:
// SlowROM
/* 00 */ .byte opcode_clocks(2, 0, 6, 0, 0) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 2, 0) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 6, 0, 0) // COP
/* 03 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 4, 0, 0) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 2, 0, 0) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 4, 0, 0) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 2, 0) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHP
/* 09 */ .byte opcode_clocks(3, 0, 0, 0, 0) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 4, 0) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 2, 0) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 4, 0) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 2, 0) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BPL r
/* 11 */ .byte opcode_clocks(2, 0, 2, 2, 0) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 2, 0) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 2, 0) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 4, 0, 0) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 4, 0, 0) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 2, 0) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLC
/* 19 */ .byte opcode_clocks(3, 0, 0, 2, 0) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 0) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 4, 0) // TRB a
/* 1D */ .byte opcode_clocks(3, 0, 0, 2, 0) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 4, 0) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 2, 0) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 0) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 2, 0) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 0) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 2, 0, 0) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 2, 0, 0) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 2, 0, 0) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 4, 0, 0) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 2, 0) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLP
/* 29 */ .byte opcode_clocks(3, 0, 0, 0, 0) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 2, 0) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 2, 0) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 4, 0) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 2, 0) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BMI r
/* 31 */ .byte opcode_clocks(2, 0, 2, 2, 0) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 2, 0) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 2, 0) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 2, 0, 0) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 2, 0, 0) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 4, 0, 0) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 2, 0) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEC
/* 39 */ .byte opcode_clocks(3, 0, 0, 2, 0) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSC
/* 3C */ .byte opcode_clocks(3, 0, 0, 2, 0) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 0, 0, 2, 0) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 4, 0) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 2, 0) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 4, 0, 0) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 2, 0) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 0) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 2, 0, 0) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 2, 0, 0) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 4, 0, 0) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 2, 0) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHA
/* 49 */ .byte opcode_clocks(3, 0, 0, 0, 0) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 0) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 2, 0) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 4, 0) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 2, 0) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVC r
/* 51 */ .byte opcode_clocks(2, 0, 2, 2, 0) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 2, 0) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 2, 0) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 2, 0, 0) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 4, 0, 0) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 2, 0) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLI
/* 59 */ .byte opcode_clocks(3, 0, 0, 2, 0) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 0) // JML al
/* 5D */ .byte opcode_clocks(3, 0, 0, 2, 0) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 4, 0) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 2, 0) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 0) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 2, 0) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 0) // PER
/* 63 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 2, 0, 0) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 2, 0, 0) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 4, 0, 0) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 2, 0) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLA
/* 69 */ .byte opcode_clocks(3, 0, 0, 0, 0) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 0) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 0) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 2, 0) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 4, 0) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 2, 0) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVS r
/* 71 */ .byte opcode_clocks(2, 0, 2, 2, 0) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 2, 0) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 2, 0) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 4, 0, 0) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 2, 0) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEI
/* 79 */ .byte opcode_clocks(3, 0, 0, 2, 0) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 0) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 0, 0, 2, 0) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 4, 0) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 2, 0) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 2, 0) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 0) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 2, 0, 0) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 2, 0) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEY
/* 89 */ .byte opcode_clocks(3, 0, 0, 0, 0) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 1, 0) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 2, 0) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 1, 0) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 2, 0) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 2, 0) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 2, 0) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 2, 0) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 2, 0, 0) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 2, 0) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 2, 0) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 2, 0) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 2, 0) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 2, 0) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 2, 0) // STA al,x

/* A0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // LDA (d,x)
/* A2 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 2, 0, 0) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAY
/* A9 */ .byte opcode_clocks(3, 0, 0, 0, 0) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 2, 0) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 2, 0) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCS r
/* B1 */ .byte opcode_clocks(2, 0, 2, 2, 0) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 2, 0) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 2, 0) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLV
/* B9 */ .byte opcode_clocks(3, 0, 0, 2, 0) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYX
/* BC */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDY a,x
/* BD */ .byte opcode_clocks(3, 0, 0, 2, 0) // LDA a,x
/* BE */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 2, 0) // LDA al,x

/* C0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 2, 0, 0) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 2, 0, 0) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 4, 0, 0) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INY
/* C9 */ .byte opcode_clocks(3, 0, 0, 0, 0) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 0) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 1, 0) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 2, 0) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 4, 0) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 2, 0) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BNE r
/* D1 */ .byte opcode_clocks(2, 0, 2, 2, 0) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 2, 0) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 2, 0) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 0) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 2, 0, 0) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 4, 0, 0) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLD
/* D9 */ .byte opcode_clocks(3, 0, 0, 2, 0) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 0) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 0) // JML (a)
/* DD */ .byte opcode_clocks(3, 0, 0, 2, 0) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 4, 0) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 2, 0) // CMP al,x

/* E0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 2, 0) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 2, 0, 0) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 2, 0, 0) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 4, 0, 0) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INX
/* E9 */ .byte opcode_clocks(3, 0, 0, 0, 0) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 0) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 0) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 1, 0) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 2, 0) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 4, 0) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 2, 0) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BEQ r
/* F1 */ .byte opcode_clocks(2, 0, 2, 2, 0) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 2, 0) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 2, 0) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 0) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 2, 0, 0) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 4, 0, 0) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 2, 0) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SED
/* F9 */ .byte opcode_clocks(3, 0, 0, 2, 0) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 0) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 0) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 0, 0, 2, 0) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 4, 0) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 2, 0) // SBC al,x

// FastROM
/* 00 */ .byte opcode_clocks(2, 0, 6, 0, 1) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 2, 1) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 6, 0, 1) // COP
/* 03 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 4, 0, 1) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 2, 0, 1) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 4, 0, 1) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 2, 1) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHP
/* 09 */ .byte opcode_clocks(3, 0, 0, 0, 1) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 4, 1) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 2, 1) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 4, 1) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 2, 1) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BPL r
/* 11 */ .byte opcode_clocks(2, 0, 2, 2, 1) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 2, 1) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 2, 1) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 4, 0, 1) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 4, 0, 1) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 2, 1) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLC
/* 19 */ .byte opcode_clocks(3, 0, 0, 2, 1) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 1) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 4, 1) // TRB a
/* 1D */ .byte opcode_clocks(3, 0, 0, 2, 1) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 4, 1) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 2, 1) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 1) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 2, 1) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 1) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 2, 0, 1) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 2, 0, 1) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 2, 0, 1) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 4, 0, 1) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 2, 1) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLP
/* 29 */ .byte opcode_clocks(3, 0, 0, 0, 1) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 2, 1) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 2, 1) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 4, 1) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 2, 1) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BMI r
/* 31 */ .byte opcode_clocks(2, 0, 2, 2, 1) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 2, 1) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 2, 1) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 2, 0, 1) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 2, 0, 1) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 4, 0, 1) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 2, 1) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEC
/* 39 */ .byte opcode_clocks(3, 0, 0, 2, 1) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSC
/* 3C */ .byte opcode_clocks(3, 0, 0, 2, 1) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 0, 0, 2, 1) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 4, 1) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 2, 1) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 4, 0, 1) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 2, 1) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 1) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 2, 0, 1) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 2, 0, 1) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 4, 0, 1) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 2, 1) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHA
/* 49 */ .byte opcode_clocks(3, 0, 0, 0, 1) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 1) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 2, 1) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 4, 1) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 2, 1) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVC r
/* 51 */ .byte opcode_clocks(2, 0, 2, 2, 1) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 2, 1) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 2, 1) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 2, 0, 1) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 4, 0, 1) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 2, 1) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLI
/* 59 */ .byte opcode_clocks(3, 0, 0, 2, 1) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 1) // JML al
/* 5D */ .byte opcode_clocks(3, 0, 0, 2, 1) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 4, 1) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 2, 1) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 1) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 2, 1) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 1) // PER
/* 63 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 2, 0, 1) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 2, 0, 1) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 4, 0, 1) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 2, 1) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLA
/* 69 */ .byte opcode_clocks(3, 0, 0, 0, 1) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 1) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 1) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 2, 1) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 4, 1) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 2, 1) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVS r
/* 71 */ .byte opcode_clocks(2, 0, 2, 2, 1) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 2, 1) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 2, 1) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 4, 0, 1) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 2, 1) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEI
/* 79 */ .byte opcode_clocks(3, 0, 0, 2, 1) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 1) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 0, 0, 2, 1) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 4, 1) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 2, 1) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 2, 1) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 1) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 2, 0, 1) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 2, 1) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEY
/* 89 */ .byte opcode_clocks(3, 0, 0, 0, 1) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 1, 1) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 2, 1) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 1, 1) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 2, 1) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 2, 1) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 2, 1) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 2, 1) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 2, 0, 1) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 2, 1) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 2, 1) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 2, 1) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 2, 1) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 2, 1) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 2, 1) // STA al,x

/* A0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // LDA (d,x)
/* A2 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 2, 0, 1) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAY
/* A9 */ .byte opcode_clocks(3, 0, 0, 0, 1) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 2, 1) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 2, 1) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCS r
/* B1 */ .byte opcode_clocks(2, 0, 2, 2, 1) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 2, 1) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 2, 1) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLV
/* B9 */ .byte opcode_clocks(3, 0, 0, 2, 1) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYX
/* BC */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDY a,x
/* BD */ .byte opcode_clocks(3, 0, 0, 2, 1) // LDA a,x
/* BE */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 2, 1) // LDA al,x

/* C0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 2, 0, 1) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 2, 0, 1) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 4, 0, 1) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INY
/* C9 */ .byte opcode_clocks(3, 0, 0, 0, 1) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 1) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 1, 1) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 2, 1) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 4, 1) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 2, 1) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BNE r
/* D1 */ .byte opcode_clocks(2, 0, 2, 2, 1) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 2, 1) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 2, 1) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 1) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 2, 0, 1) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 4, 0, 1) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLD
/* D9 */ .byte opcode_clocks(3, 0, 0, 2, 1) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 1) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 1) // JML (a)
/* DD */ .byte opcode_clocks(3, 0, 0, 2, 1) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 4, 1) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 2, 1) // CMP al,x

/* E0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 2, 1) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 2, 0, 1) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 2, 0, 1) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 4, 0, 1) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INX
/* E9 */ .byte opcode_clocks(3, 0, 0, 0, 1) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 1) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 1) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 1, 1) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 2, 1) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 4, 1) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 2, 1) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BEQ r
/* F1 */ .byte opcode_clocks(2, 0, 2, 2, 1) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 2, 1) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 2, 1) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 1) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 2, 0, 1) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 4, 0, 1) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 2, 1) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SED
/* F9 */ .byte opcode_clocks(3, 0, 0, 2, 1) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 1) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 1) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 0, 0, 2, 1) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 4, 1) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 2, 1) // SBC al,x

OpTableMX:
/* 00 */ .long   E0_BRK       ,EM_ORA_Od_xO,E0_COP    ,EM_ORA_d_s
/* 04 */ .long   EM_TSB_d     ,EM_ORA_d    ,EM_ASL_d  ,EM_ORA_IdI
/* 08 */ .long   E0_PHP       ,EM_ORA_i    ,EM_SLA    ,E0_PHD
/* 0C */ .long   EM_TSB_a     ,EM_ORA_a    ,EM_ASL_a  ,EM_ORA_al
/* 10 */ .long   ALL_BPL      ,EM_ORA_OdO_y,EM_ORA_OdO,EM_ORA_Od_sO_y
/* 14 */ .long   EM_TRB_d     ,EM_ORA_d_x  ,EM_ASL_d_x,EM_ORA_IdI_y
/* 18 */ .long   ALL_CLC      ,EM_ORA_a_y  ,EM_INA    ,E0_TCS
/* 1C */ .long   EM_TRB_a     ,EM_ORA_a_x  ,EM_ASL_a_x,EM_ORA_al_x
/* 20 */ .long   E0_JSR_a     ,EM_AND_Od_xO,E0_JSL_al ,EM_AND_d_s
/* 24 */ .long   EM_BIT_d     ,EM_AND_d    ,EM_ROL_d  ,EM_AND_IdI
/* 28 */ .long   E0_PLP       ,EM_AND_i    ,EM_RLA    ,E0_PLD
/* 2C */ .long   EM_BIT_a     ,EM_AND_a    ,EM_ROL_a  ,EM_AND_al
/* 30 */ .long   ALL_BMI      ,EM_AND_OdO_y,EM_AND_OdO,EM_AND_Od_sO_y
/* 34 */ .long   EM_BIT_d_x   ,EM_AND_d_x  ,EM_ROL_d_x,EM_AND_IdI_y
/* 38 */ .long   ALL_SEC      ,EM_AND_a_y  ,EM_DEA    ,ALL_TSC
/* 3C */ .long   EM_BIT_a_x   ,EM_AND_a_x  ,EM_ROL_a_x,EM_AND_al_x
/* 40 */ .long   E0_RTI       ,EM_EOR_Od_xO,DO_INVALID,EM_EOR_d_s
/* 44 */ .long   EX_MVP       ,EM_EOR_d    ,EM_LSR_d  ,EM_EOR_IdI
/* 48 */ .long   EM_PHA       ,EM_EOR_i    ,EM_SRA    ,E0_PHK
/* 4C */ .long   ALL_JMP_a    ,EM_EOR_a    ,EM_LSR_a  ,EM_EOR_al
/* 50 */ .long   ALL_BVC      ,EM_EOR_OdO_y,EM_EOR_OdO,EM_EOR_Od_sO_y
/* 54 */ .long   EX_MVN       ,EM_EOR_d_x  ,EM_LSR_d_x,EM_EOR_IdI_y
/* 58 */ .long   E0_CLI       ,EM_EOR_a_y  ,EX_PHY    ,ALL_TCD
/* 5C */ .long   ALL_JML_al   ,EM_EOR_a_x  ,EM_LSR_a_x,EM_EOR_al_x
/* 60 */ .long   E0_RTS       ,EM_ADC_Od_xO,E0_PER    ,EM_ADC_d_s
/* 64 */ .long   EM_STZ_d     ,EM_ADC_d    ,EM_ROR_d  ,EM_ADC_IdI
/* 68 */ .long   EM_PLA       ,EM_ADC_i    ,EM_RRA    ,E0_RTL
/* 6C */ .long   ALL_JMP_OaO  ,EM_ADC_a    ,EM_ROR_a  ,EM_ADC_al
/* 70 */ .long   ALL_BVS      ,EM_ADC_OdO_y,EM_ADC_OdO,EM_ADC_Od_sO_y
/* 74 */ .long   EM_STZ_d_x   ,EM_ADC_d_x  ,EM_ROR_d_x,EM_ADC_IdI_y
/* 78 */ .long   ALL_SEI      ,EM_ADC_a_y  ,EX_PLY    ,ALL_TDC
/* 7C */ .long   ALL_JMP_Oa_xO,EM_ADC_a_x  ,EM_ROR_a_x,EM_ADC_al_x
/* 80 */ .long   ALL_BRA      ,EM_STA_Od_xO,ALL_BRL   ,EM_STA_d_s
/* 84 */ .long   EX_STY_d     ,EM_STA_d    ,EX_STX_d  ,EM_STA_IdI
/* 88 */ .long   EX_DEY       ,EM_BIT_i    ,EM_TXA    ,E0_PHB
/* 8C */ .long   EX_STY_a     ,EM_STA_a    ,EX_STX_a  ,EM_STA_al
/* 90 */ .long   ALL_BCC      ,EM_STA_OdO_y,EM_STA_OdO,EM_STA_Od_sO_y
/* 94 */ .long   EX_STY_d_x   ,EM_STA_d_x  ,EX_STX_d_y,EM_STA_IdI_y
/* 98 */ .long   EM_TYA       ,EM_STA_a_y  ,E0_TXS    ,EX_TXY
/* 9C */ .long   EM_STZ_a     ,EM_STA_a_x  ,EM_STZ_a_x,EM_STA_al_x
/* A0 */ .long   EX_LDY_i     ,EM_LDA_Od_xO,EX_LDX_i  ,EM_LDA_d_s
/* A4 */ .long   EX_LDY_d     ,EM_LDA_d    ,EX_LDX_d  ,EM_LDA_IdI
/* A8 */ .long   EX_TAY       ,EM_LDA_i    ,EX_TAX    ,E0_PLB
/* AC */ .long   EX_LDY_a     ,EM_LDA_a    ,EX_LDX_a  ,EM_LDA_al
/* B0 */ .long   ALL_BCS      ,EM_LDA_OdO_y,EM_LDA_OdO,EM_LDA_Od_sO_y
/* B4 */ .long   EX_LDY_d_x   ,EM_LDA_d_x  ,EX_LDX_d_y,EM_LDA_IdI_y
/* B8 */ .long   ALL_CLV      ,EM_LDA_a_y  ,EX_TSX    ,EX_TYX
/* BC */ .long   EX_LDY_a_x   ,EM_LDA_a_x  ,EX_LDX_a_y,EM_LDA_al_x
/* C0 */ .long   EX_CPY_i     ,EM_CMP_Od_xO,E0_REP    ,EM_CMP_d_s
/* C4 */ .long   EX_CPY_d     ,EM_CMP_d    ,EM_DEC_d  ,EM_CMP_IdI
/* C8 */ .long   EX_INY       ,EM_CMP_i    ,EX_DEX    ,ALL_WAI
/* CC */ .long   EX_CPY_a     ,EM_CMP_a    ,EM_DEC_a  ,EM_CMP_al
/* D0 */ .long   ALL_BNE      ,EM_CMP_OdO_y,EM_CMP_OdO,EM_CMP_Od_sO_y
/* D4 */ .long   E0_PEI       ,EM_CMP_d_x  ,EM_DEC_d_x,EM_CMP_IdI_y
/* D8 */ .long   ALL_CLD      ,EM_CMP_a_y  ,EX_PHX    ,DO_INVALID
/* DC */ .long   ALL_JML_OaO  ,EM_CMP_a_x  ,EM_DEC_a_x,EM_CMP_al_x
/* E0 */ .long   EX_CPX_i     ,EM_SBC_Od_xO,E0_SEP    ,EM_SBC_d_s
/* E4 */ .long   EX_CPX_d     ,EM_SBC_d    ,EM_INC_d  ,EM_SBC_IdI
/* E8 */ .long   EX_INX       ,EM_SBC_i    ,ALL_NOP   ,ALL_XBA
/* EC */ .long   EX_CPX_a     ,EM_SBC_a    ,EM_INC_a  ,EM_SBC_al
/* F0 */ .long   ALL_BEQ      ,EM_SBC_OdO_y,EM_SBC_OdO,EM_SBC_Od_sO_y
/* F4 */ .long   E0_PEA       ,EM_SBC_d_x  ,EM_INC_d_x,EM_SBC_IdI_y
/* F8 */ .long   ALL_SED      ,EM_SBC_a_y  ,EX_PLX    ,E0_XCE
/* FC */ .long   E0_JSR_Oa_xO ,EM_SBC_a_x  ,EM_INC_a_x,EM_SBC_al_x

// bytes, internal operations, bank 0 accesses, other bus accesses, speed
//  speed = 0 for SlowROM, 1 for FastROM
CCTableMX:
// SlowROM
/* 00 */ .byte opcode_clocks(2, 0, 6, 0, 0) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 1, 0) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 6, 0, 0) // COP
/* 03 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 2, 0, 0) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 1, 0, 0) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHP
/* 09 */ .byte opcode_clocks(2, 0, 0, 0, 0) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 2, 0) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BPL r
/* 11 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 1, 0) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 2, 0, 0) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLC
/* 19 */ .byte opcode_clocks(3, 0, 0, 1, 0) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 0) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 2, 0) // TRB a
/* 1D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 0) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 1, 0) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 0) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 1, 0, 0) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 1, 0, 0) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 1, 0, 0) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 1, 0) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLP
/* 29 */ .byte opcode_clocks(2, 0, 0, 0, 0) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 1, 0) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 1, 0) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 1, 0) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BMI r
/* 31 */ .byte opcode_clocks(2, 0, 2, 1, 0) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 1, 0) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 1, 0) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 1, 0, 0) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 1, 0, 0) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 1, 0) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEC
/* 39 */ .byte opcode_clocks(3, 0, 0, 1, 0) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSC
/* 3C */ .byte opcode_clocks(3, 0, 0, 1, 0) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 0, 0, 1, 0) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 1, 0) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 4, 0, 0) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 1, 0) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 0) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 1, 0, 0) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 1, 0, 0) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 1, 0) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHA
/* 49 */ .byte opcode_clocks(2, 0, 0, 0, 0) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 0) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 1, 0) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 2, 0) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 1, 0) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVC r
/* 51 */ .byte opcode_clocks(2, 0, 2, 1, 0) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 1, 0) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 1, 0) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 1, 0, 0) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 2, 0, 0) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 1, 0) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLI
/* 59 */ .byte opcode_clocks(3, 0, 0, 1, 0) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 0) // JML al
/* 5D */ .byte opcode_clocks(3, 0, 0, 1, 0) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 2, 0) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 1, 0) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 0) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 1, 0) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 0) // PER
/* 63 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 1, 0, 0) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLA
/* 69 */ .byte opcode_clocks(2, 0, 0, 0, 0) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 0) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 0) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVS r
/* 71 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 1, 0) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEI
/* 79 */ .byte opcode_clocks(3, 0, 0, 1, 0) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 0) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 1, 0) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 0) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 1, 0) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEY
/* 89 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 1, 0) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 1, 0) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 1, 0) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 1, 0) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 1, 0) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 1, 0) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 1, 0) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 1, 0) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 1, 0) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 1, 0) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 1, 0) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 1, 0) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 1, 0) // STA al,x

/* A0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // LDA (d,x)
/* A2 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAY
/* A9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 1, 0) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCS r
/* B1 */ .byte opcode_clocks(2, 0, 2, 1, 0) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLV
/* B9 */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYX
/* BC */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDY a,x
/* BD */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDA a,x
/* BE */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 1, 0) // LDA al,x

/* C0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 2, 0, 0) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INY
/* C9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 0) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 1, 0) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 1, 0) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 2, 0) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 1, 0) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BNE r
/* D1 */ .byte opcode_clocks(2, 0, 2, 1, 0) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 0) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 2, 0, 0) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLD
/* D9 */ .byte opcode_clocks(3, 0, 0, 1, 0) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 0) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 0) // JML (a)
/* DD */ .byte opcode_clocks(3, 0, 0, 1, 0) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 2, 0) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 1, 0) // CMP al,x

/* E0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 2, 0, 0) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INX
/* E9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 0) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 0) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 1, 0) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 1, 0) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 2, 0) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 1, 0) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BEQ r
/* F1 */ .byte opcode_clocks(2, 0, 2, 1, 0) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 0) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 2, 0, 0) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SED
/* F9 */ .byte opcode_clocks(3, 0, 0, 1, 0) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 0) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 0) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 0, 0, 1, 0) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 2, 0) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 1, 0) // SBC al,x

// FastROM
/* 00 */ .byte opcode_clocks(2, 0, 6, 0, 1) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 1, 1) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 6, 0, 1) // COP
/* 03 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 2, 0, 1) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 1, 0, 1) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHP
/* 09 */ .byte opcode_clocks(2, 0, 0, 0, 1) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 2, 1) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BPL r
/* 11 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 1, 1) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 2, 0, 1) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLC
/* 19 */ .byte opcode_clocks(3, 0, 0, 1, 1) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 1) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 2, 1) // TRB a
/* 1D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 1) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 1, 1) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 1) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 1, 0, 1) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 1, 0, 1) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 1, 0, 1) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 1, 1) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLP
/* 29 */ .byte opcode_clocks(2, 0, 0, 0, 1) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 1, 1) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 1, 1) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 1, 1) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BMI r
/* 31 */ .byte opcode_clocks(2, 0, 2, 1, 1) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 1, 1) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 1, 1) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 1, 0, 1) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 1, 0, 1) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 1, 1) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEC
/* 39 */ .byte opcode_clocks(3, 0, 0, 1, 1) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSC
/* 3C */ .byte opcode_clocks(3, 0, 0, 1, 1) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 0, 0, 1, 1) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 1, 1) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 4, 0, 1) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 1, 1) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 1) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 1, 0, 1) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 1, 0, 1) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 1, 1) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHA
/* 49 */ .byte opcode_clocks(2, 0, 0, 0, 1) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 1) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 1, 1) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 2, 1) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 1, 1) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVC r
/* 51 */ .byte opcode_clocks(2, 0, 2, 1, 1) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 1, 1) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 1, 1) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 1, 0, 1) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 2, 0, 1) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 1, 1) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLI
/* 59 */ .byte opcode_clocks(3, 0, 0, 1, 1) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 1) // JML al
/* 5D */ .byte opcode_clocks(3, 0, 0, 1, 1) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 2, 1) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 1, 1) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 1) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 1, 1) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 1) // PER
/* 63 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 1, 0, 1) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLA
/* 69 */ .byte opcode_clocks(2, 0, 0, 0, 1) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 1) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 1) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVS r
/* 71 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 1, 1) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEI
/* 79 */ .byte opcode_clocks(3, 0, 0, 1, 1) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 1) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 1, 1) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 1) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 1, 1) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEY
/* 89 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 1, 1) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 1, 1) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 1, 1) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 1, 1) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 1, 1) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 1, 1) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 1, 1) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 1, 1) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 1, 1) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 1, 1) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 1, 1) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 1, 1) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 1, 1) // STA al,x

/* A0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // LDA (d,x)
/* A2 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAY
/* A9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 1, 1) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCS r
/* B1 */ .byte opcode_clocks(2, 0, 2, 1, 1) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLV
/* B9 */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYX
/* BC */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDY a,x
/* BD */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDA a,x
/* BE */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 1, 1) // LDA al,x

/* C0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 2, 0, 1) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INY
/* C9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 1) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 1, 1) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 1, 1) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 2, 1) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 1, 1) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BNE r
/* D1 */ .byte opcode_clocks(2, 0, 2, 1, 1) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 1) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 2, 0, 1) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLD
/* D9 */ .byte opcode_clocks(3, 0, 0, 1, 1) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 1) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 1) // JML (a)
/* DD */ .byte opcode_clocks(3, 0, 0, 1, 1) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 2, 1) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 1, 1) // CMP al,x

/* E0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 2, 0, 1) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INX
/* E9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 1) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 1) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 1, 1) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 1, 1) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 2, 1) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 1, 1) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BEQ r
/* F1 */ .byte opcode_clocks(2, 0, 2, 1, 1) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 1) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 2, 0, 1) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SED
/* F9 */ .byte opcode_clocks(3, 0, 0, 1, 1) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 1) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 1) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 0, 0, 1, 1) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 2, 1) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 1, 1) // SBC al,x

OpTableE1:
/* 00 */ .long   E1_BRK       ,EM_ORA_Od_xO,E1_COP    ,EM_ORA_d_s
/* 04 */ .long   EM_TSB_d     ,EM_ORA_d    ,EM_ASL_d  ,EM_ORA_IdI
/* 08 */ .long   E1_PHP       ,EM_ORA_i    ,EM_SLA    ,E1_PHD
/* 0C */ .long   EM_TSB_a     ,EM_ORA_a    ,EM_ASL_a  ,EM_ORA_al
/* 10 */ .long   ALL_BPL      ,EM_ORA_OdO_y,EM_ORA_OdO,EM_ORA_Od_sO_y
/* 14 */ .long   EM_TRB_d     ,EM_ORA_d_x  ,EM_ASL_d_x,EM_ORA_IdI_y
/* 18 */ .long   ALL_CLC      ,EM_ORA_a_y  ,EM_INA    ,E1_TCS
/* 1C */ .long   EM_TRB_a     ,EM_ORA_a_x  ,EM_ASL_a_x,EM_ORA_al_x
/* 20 */ .long   E1_JSR_a     ,EM_AND_Od_xO,E1_JSL_al ,EM_AND_d_s
/* 24 */ .long   EM_BIT_d     ,EM_AND_d    ,EM_ROL_d  ,EM_AND_IdI
/* 28 */ .long   E1_PLP       ,EM_AND_i    ,EM_RLA    ,E1_PLD
/* 2C */ .long   EM_BIT_a     ,EM_AND_a    ,EM_ROL_a  ,EM_AND_al
/* 30 */ .long   ALL_BMI      ,EM_AND_OdO_y,EM_AND_OdO,EM_AND_Od_sO_y
/* 34 */ .long   EM_BIT_d_x   ,EM_AND_d_x  ,EM_ROL_d_x,EM_AND_IdI_y
/* 38 */ .long   ALL_SEC      ,EM_AND_a_y  ,EM_DEA    ,ALL_TSC
/* 3C */ .long   EM_BIT_a_x   ,EM_AND_a_x  ,EM_ROL_a_x,EM_AND_al_x
/* 40 */ .long   E1_RTI       ,EM_EOR_Od_xO,DO_INVALID,EM_EOR_d_s
/* 44 */ .long   EX_MVP       ,EM_EOR_d    ,EM_LSR_d  ,EM_EOR_IdI
/* 48 */ .long   EM_PHA       ,EM_EOR_i    ,EM_SRA    ,E1_PHK
/* 4C */ .long   ALL_JMP_a    ,EM_EOR_a    ,EM_LSR_a  ,EM_EOR_al
/* 50 */ .long   ALL_BVC      ,EM_EOR_OdO_y,EM_EOR_OdO,EM_EOR_Od_sO_y
/* 54 */ .long   EX_MVN       ,EM_EOR_d_x  ,EM_LSR_d_x,EM_EOR_IdI_y
/* 58 */ .long   E1_CLI       ,EM_EOR_a_y  ,E1_PHY    ,ALL_TCD
/* 5C */ .long   ALL_JML_al   ,EM_EOR_a_x  ,EM_LSR_a_x,EM_EOR_al_x
/* 60 */ .long   E1_RTS       ,EM_ADC_Od_xO,E1_PER    ,EM_ADC_d_s
/* 64 */ .long   EM_STZ_d     ,EM_ADC_d    ,EM_ROR_d  ,EM_ADC_IdI
/* 68 */ .long   E1_PLA       ,EM_ADC_i    ,EM_RRA    ,E1_RTL
/* 6C */ .long   ALL_JMP_OaO  ,EM_ADC_a    ,EM_ROR_a  ,EM_ADC_al
/* 70 */ .long   ALL_BVS      ,EM_ADC_OdO_y,EM_ADC_OdO,EM_ADC_Od_sO_y
/* 74 */ .long   EM_STZ_d_x   ,EM_ADC_d_x  ,EM_ROR_d_x,EM_ADC_IdI_y
/* 78 */ .long   ALL_SEI      ,EM_ADC_a_y  ,E1_PLY    ,ALL_TDC
/* 7C */ .long   ALL_JMP_Oa_xO,EM_ADC_a_x  ,EM_ROR_a_x,EM_ADC_al_x
/* 80 */ .long   ALL_BRA      ,EM_STA_Od_xO,ALL_BRL   ,EM_STA_d_s
/* 84 */ .long   EX_STY_d     ,EM_STA_d    ,EX_STX_d  ,EM_STA_IdI
/* 88 */ .long   EX_DEY       ,EM_BIT_i    ,EM_TXA    ,E1_PHB
/* 8C */ .long   EX_STY_a     ,EM_STA_a    ,EX_STX_a  ,EM_STA_al
/* 90 */ .long   ALL_BCC      ,EM_STA_OdO_y,EM_STA_OdO,EM_STA_Od_sO_y
/* 94 */ .long   EX_STY_d_x   ,EM_STA_d_x  ,EX_STX_d_y,EM_STA_IdI_y
/* 98 */ .long   EM_TYA       ,EM_STA_a_y  ,E1_TXS    ,EX_TXY
/* 9C */ .long   EM_STZ_a     ,EM_STA_a_x  ,EM_STZ_a_x,EM_STA_al_x
/* A0 */ .long   EX_LDY_i     ,EM_LDA_Od_xO,EX_LDX_i  ,EM_LDA_d_s
/* A4 */ .long   EX_LDY_d     ,EM_LDA_d    ,EX_LDX_d  ,EM_LDA_IdI
/* A8 */ .long   EX_TAY       ,EM_LDA_i    ,EX_TAX    ,E1_PLB
/* AC */ .long   EX_LDY_a     ,EM_LDA_a    ,EX_LDX_a  ,EM_LDA_al
/* B0 */ .long   ALL_BCS      ,EM_LDA_OdO_y,EM_LDA_OdO,EM_LDA_Od_sO_y
/* B4 */ .long   EX_LDY_d_x   ,EM_LDA_d_x  ,EX_LDX_d_y,EM_LDA_IdI_y
/* B8 */ .long   ALL_CLV      ,EM_LDA_a_y  ,EX_TSX    ,EX_TYX
/* BC */ .long   EX_LDY_a_x   ,EM_LDA_a_x  ,EX_LDX_a_y,EM_LDA_al_x
/* C0 */ .long   EX_CPY_i     ,EM_CMP_Od_xO,E1_REP    ,EM_CMP_d_s
/* C4 */ .long   EX_CPY_d     ,EM_CMP_d    ,EM_DEC_d  ,EM_CMP_IdI
/* C8 */ .long   EX_INY       ,EM_CMP_i    ,EX_DEX    ,ALL_WAI
/* CC */ .long   EX_CPY_a     ,EM_CMP_a    ,EM_DEC_a  ,EM_CMP_al
/* D0 */ .long   ALL_BNE      ,EM_CMP_OdO_y,EM_CMP_OdO,EM_CMP_Od_sO_y
/* D4 */ .long   E1_PEI       ,EM_CMP_d_x  ,EM_DEC_d_x,EM_CMP_IdI_y
/* D8 */ .long   ALL_CLD      ,EM_CMP_a_y  ,E1_PHX    ,DO_INVALID
/* DC */ .long   ALL_JML_OaO  ,EM_CMP_a_x  ,EM_DEC_a_x,EM_CMP_al_x
/* E0 */ .long   EX_CPX_i     ,EM_SBC_Od_xO,E1_SEP    ,EM_SBC_d_s
/* E4 */ .long   EX_CPX_d     ,EM_SBC_d    ,EM_INC_d  ,EM_SBC_IdI
/* E8 */ .long   EX_INX       ,EM_SBC_i    ,ALL_NOP   ,ALL_XBA
/* EC */ .long   EX_CPX_a     ,EM_SBC_a    ,EM_INC_a  ,EM_SBC_al
/* F0 */ .long   ALL_BEQ      ,EM_SBC_OdO_y,EM_SBC_OdO,EM_SBC_Od_sO_y
/* F4 */ .long   E1_PEA       ,EM_SBC_d_x  ,EM_INC_d_x,EM_SBC_IdI_y
/* F8 */ .long   ALL_SED      ,EM_SBC_a_y  ,E1_PLX    ,E1_XCE
/* FC */ .long   E1_JSR_Oa_xO ,EM_SBC_a_x  ,EM_INC_a_x,EM_SBC_al_x

// bytes, internal operations, bank 0 accesses, other bus accesses, speed
//  speed = 0 for SlowROM, 1 for FastROM
CCTableE1:
// SlowROM
/* 00 */ .byte opcode_clocks(2, 0, 5, 0, 0) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 1, 0) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 5, 0, 0) // COP
/* 03 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 2, 0, 0) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 1, 0, 0) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHP
/* 09 */ .byte opcode_clocks(2, 0, 0, 0, 0) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 0) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 2, 0) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BPL r
/* 11 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 1, 0) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 2, 0, 0) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLC
/* 19 */ .byte opcode_clocks(3, 0, 0, 1, 0) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 0) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 2, 0) // TRB a
/* 1D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 0) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 1, 0) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 0) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 1, 0, 0) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 1, 0, 0) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 1, 0, 0) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 1, 0) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLP
/* 29 */ .byte opcode_clocks(2, 0, 0, 0, 0) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 0) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 1, 0) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 1, 0) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 1, 0) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BMI r
/* 31 */ .byte opcode_clocks(2, 0, 2, 1, 0) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 1, 0) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 1, 0) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 1, 0, 0) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 1, 0, 0) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 1, 0) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEC
/* 39 */ .byte opcode_clocks(3, 0, 0, 1, 0) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSC
/* 3C */ .byte opcode_clocks(3, 0, 0, 1, 0) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 0, 0, 1, 0) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 1, 0) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 3, 0, 0) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 1, 0) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 0) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 1, 0, 0) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 1, 0, 0) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 2, 0, 0) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 1, 0) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHA
/* 49 */ .byte opcode_clocks(2, 0, 0, 0, 0) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 0) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 0) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 1, 0) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 2, 0) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 1, 0) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVC r
/* 51 */ .byte opcode_clocks(2, 0, 2, 1, 0) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 1, 0) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 1, 0) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 0) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 1, 0, 0) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 2, 0, 0) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 1, 0) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLI
/* 59 */ .byte opcode_clocks(3, 0, 0, 1, 0) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 0) // JML al
/* 5D */ .byte opcode_clocks(3, 0, 0, 1, 0) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 2, 0) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 1, 0) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 0) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 1, 0) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 0) // PER
/* 63 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 1, 0, 0) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 2, 0, 0) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLA
/* 69 */ .byte opcode_clocks(2, 0, 0, 0, 0) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 0) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 0) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 0) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 2, 0) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BVS r
/* 71 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 1, 0) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 1, 0) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 1, 0, 0) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 2, 0, 0) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 1, 0) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SEI
/* 79 */ .byte opcode_clocks(3, 0, 0, 1, 0) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 0) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 0, 0, 1, 0) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 2, 0) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 1, 0) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 1, 0) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 0) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 1, 0, 0) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 1, 0) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEY
/* 89 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 1, 0) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 1, 0) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 1, 0) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 1, 0) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 1, 0) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 1, 0) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 1, 0) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 1, 0, 0) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 1, 0) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 1, 0) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 0) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 1, 0) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 1, 0) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 1, 0) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 1, 0) // STA al,x

/* A0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // LDA (d,x)
/* A2 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 1, 0, 0) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAY
/* A9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 1, 0) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BCS r
/* B1 */ .byte opcode_clocks(2, 0, 2, 1, 0) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 1, 0, 0) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLV
/* B9 */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 0) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 0) // TYX
/* BC */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDY a,x
/* BD */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDA a,x
/* BE */ .byte opcode_clocks(3, 0, 0, 1, 0) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 1, 0) // LDA al,x

/* C0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 2, 0, 0) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INY
/* C9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 0) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 0) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 1, 0) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 1, 0) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 2, 0) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 1, 0) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BNE r
/* D1 */ .byte opcode_clocks(2, 0, 2, 1, 0) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 0) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 2, 0, 0) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // CLD
/* D9 */ .byte opcode_clocks(3, 0, 0, 1, 0) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 1, 0, 0) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 0) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 0) // JML (a)
/* DD */ .byte opcode_clocks(3, 0, 0, 1, 0) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 2, 0) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 1, 0) // CMP al,x

/* E0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 1, 0) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 0) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 1, 0, 0) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 1, 0, 0) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 1, 0, 0) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 2, 0, 0) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // INX
/* E9 */ .byte opcode_clocks(2, 0, 0, 0, 0) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 0) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 0) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 1, 0) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 1, 0) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 2, 0) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 1, 0) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 0) // BEQ r
/* F1 */ .byte opcode_clocks(2, 0, 2, 1, 0) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 1, 0) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 1, 0) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 0) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 1, 0, 0) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 2, 0, 0) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 1, 0) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 0) // SED
/* F9 */ .byte opcode_clocks(3, 0, 0, 1, 0) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 1, 0, 0) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 0) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 0) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 0, 0, 1, 0) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 2, 0) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 1, 0) // SBC al,x

// FastROM
/* 00 */ .byte opcode_clocks(2, 0, 5, 0, 1) // BRK
/* 01 */ .byte opcode_clocks(2, 1, 2, 1, 1) // ORA (d,x)
/* 02 */ .byte opcode_clocks(2, 0, 5, 0, 1) // COP
/* 03 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ORA d,s
/* 04 */ .byte opcode_clocks(2, 1, 2, 0, 1) // TSB d
/* 05 */ .byte opcode_clocks(2, 0, 1, 0, 1) // ORA d
/* 06 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ASL d
/* 07 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ORA [d]

/* 08 */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHP
/* 09 */ .byte opcode_clocks(2, 0, 0, 0, 1) // ORA i
/* 0A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SLA
/* 0B */ .byte opcode_clocks(1, 1, 2, 0, 1) // PHD
/* 0C */ .byte opcode_clocks(3, 1, 0, 2, 1) // TSB a
/* 0D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ORA a
/* 0E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ASL a
/* 0F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ORA al

/* 10 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BPL r
/* 11 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ORA (d),y
/* 12 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ORA (d)
/* 13 */ .byte opcode_clocks(2, 2, 2, 1, 1) // ORA (d,s),y
/* 14 */ .byte opcode_clocks(2, 1, 2, 0, 1) // TRB d
/* 15 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ORA d,x
/* 16 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ASL d,x
/* 17 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ORA [d],y

/* 18 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLC
/* 19 */ .byte opcode_clocks(3, 0, 0, 1, 1) // ORA a,y
/* 1A */ .byte opcode_clocks(1, 1, 0, 0, 1) // INA
/* 1B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCS
/* 1C */ .byte opcode_clocks(3, 1, 0, 2, 1) // TRB a
/* 1D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ORA a,x
/* 1E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ASL a,x
/* 1F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ORA al,x

/* 20 */ .byte opcode_clocks(3, 1, 2, 0, 1) // JSR a
/* 21 */ .byte opcode_clocks(2, 1, 2, 1, 1) // AND (d,x)
/* 22 */ .byte opcode_clocks(4, 1, 3, 0, 1) // JSL al
/* 23 */ .byte opcode_clocks(2, 1, 1, 0, 1) // AND d,s
/* 24 */ .byte opcode_clocks(2, 0, 1, 0, 1) // BIT d
/* 25 */ .byte opcode_clocks(2, 0, 1, 0, 1) // AND d
/* 26 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ROL d
/* 27 */ .byte opcode_clocks(2, 0, 3, 1, 1) // AND [d]

/* 28 */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLP
/* 29 */ .byte opcode_clocks(2, 0, 0, 0, 1) // AND i
/* 2A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RLA
/* 2B */ .byte opcode_clocks(1, 2, 2, 0, 1) // PLD
/* 2C */ .byte opcode_clocks(3, 0, 0, 1, 1) // BIT a
/* 2D */ .byte opcode_clocks(3, 0, 0, 1, 1) // AND a
/* 2E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ROL a
/* 2F */ .byte opcode_clocks(4, 0, 0, 1, 1) // AND al

/* 30 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BMI r
/* 31 */ .byte opcode_clocks(2, 0, 2, 1, 1) // AND (d),y
/* 32 */ .byte opcode_clocks(2, 0, 2, 1, 1) // AND (d)
/* 33 */ .byte opcode_clocks(2, 2, 2, 1, 1) // AND (d,s),y
/* 34 */ .byte opcode_clocks(2, 1, 1, 0, 1) // BIT d,x
/* 35 */ .byte opcode_clocks(2, 1, 1, 0, 1) // AND d,x
/* 36 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ROL d,x
/* 37 */ .byte opcode_clocks(2, 0, 3, 1, 1) // AND [d],y

/* 38 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEC
/* 39 */ .byte opcode_clocks(3, 0, 0, 1, 1) // AND a,y
/* 3A */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEA
/* 3B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSC
/* 3C */ .byte opcode_clocks(3, 0, 0, 1, 1) // BIT a,x
/* 3D */ .byte opcode_clocks(3, 0, 0, 1, 1) // AND a,x
/* 3E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ROL a,x
/* 3F */ .byte opcode_clocks(4, 0, 0, 1, 1) // AND al,x

/* 40 */ .byte opcode_clocks(1, 2, 3, 0, 1) // RTI
/* 41 */ .byte opcode_clocks(2, 1, 2, 1, 1) // EOR (d,x)
/* 42 */ .byte opcode_clocks(2, 0, 0, 0, 1) // WDM *
/* 43 */ .byte opcode_clocks(2, 1, 1, 0, 1) // EOR d,s
/* 44 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVP
/* 45 */ .byte opcode_clocks(2, 0, 1, 0, 1) // EOR d
/* 46 */ .byte opcode_clocks(2, 1, 2, 0, 1) // LSR d
/* 47 */ .byte opcode_clocks(2, 0, 3, 1, 1) // EOR [d]

/* 48 */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHA
/* 49 */ .byte opcode_clocks(2, 0, 0, 0, 1) // EOR i
/* 4A */ .byte opcode_clocks(1, 1, 0, 0, 1) // SRA
/* 4B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHK
/* 4C */ .byte opcode_clocks(3, 0, 0, 0, 1) // JMP a
/* 4D */ .byte opcode_clocks(3, 0, 0, 1, 1) // EOR a
/* 4E */ .byte opcode_clocks(3, 1, 0, 2, 1) // LSR a
/* 4F */ .byte opcode_clocks(4, 0, 0, 1, 1) // EOR al

/* 50 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVC r
/* 51 */ .byte opcode_clocks(2, 0, 2, 1, 1) // EOR (d),y
/* 52 */ .byte opcode_clocks(2, 0, 2, 1, 1) // EOR (d)
/* 53 */ .byte opcode_clocks(2, 2, 2, 1, 1) // EOR (d,s),y
/* 54 */ .byte opcode_clocks(3, 2, 0, 2, 1) // MVN
/* 55 */ .byte opcode_clocks(2, 1, 1, 0, 1) // EOR d,x
/* 56 */ .byte opcode_clocks(2, 2, 2, 0, 1) // LSR d,x
/* 57 */ .byte opcode_clocks(2, 0, 3, 1, 1) // EOR [d],y

/* 58 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLI
/* 59 */ .byte opcode_clocks(3, 0, 0, 1, 1) // EOR a,y
/* 5A */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHY
/* 5B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TCD
/* 5C */ .byte opcode_clocks(4, 0, 0, 0, 1) // JML al
/* 5D */ .byte opcode_clocks(3, 0, 0, 1, 1) // EOR a,x
/* 5E */ .byte opcode_clocks(3, 2, 0, 2, 1) // LSR a,x
/* 5F */ .byte opcode_clocks(4, 0, 0, 1, 1) // EOR al,x

/* 60 */ .byte opcode_clocks(1, 3, 2, 0, 1) // RTS
/* 61 */ .byte opcode_clocks(2, 1, 2, 1, 1) // ADC (d,x)
/* 62 */ .byte opcode_clocks(3, 1, 2, 0, 1) // PER
/* 63 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ADC d,s
/* 64 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STZ d
/* 65 */ .byte opcode_clocks(2, 0, 1, 0, 1) // ADC d
/* 66 */ .byte opcode_clocks(2, 1, 2, 0, 1) // ROR d
/* 67 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ADC [d]

/* 68 */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLA
/* 69 */ .byte opcode_clocks(2, 0, 0, 0, 1) // ADC i
/* 6A */ .byte opcode_clocks(1, 1, 0, 0, 1) // RRA
/* 6B */ .byte opcode_clocks(1, 2, 3, 0, 1) // RTL
/* 6C */ .byte opcode_clocks(3, 0, 2, 0, 1) // JMP (a)
/* 6D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ADC a
/* 6E */ .byte opcode_clocks(3, 1, 0, 2, 1) // ROR a
/* 6F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ADC al

/* 70 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BVS r
/* 71 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ADC (d),y
/* 72 */ .byte opcode_clocks(2, 0, 2, 1, 1) // ADC (d)
/* 73 */ .byte opcode_clocks(2, 2, 2, 1, 1) // ADC (d,s),y
/* 74 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STZ d,x
/* 75 */ .byte opcode_clocks(2, 1, 1, 0, 1) // ADC d,x
/* 76 */ .byte opcode_clocks(2, 2, 2, 0, 1) // ROR d,x
/* 77 */ .byte opcode_clocks(2, 0, 3, 1, 1) // ADC [d],y

/* 78 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SEI
/* 79 */ .byte opcode_clocks(3, 0, 0, 1, 1) // ADC a,y
/* 7A */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLY
/* 7B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TDC
/* 7C */ .byte opcode_clocks(3, 1, 0, 2, 1) // JMP (a,x) - bus access in PB
/* 7D */ .byte opcode_clocks(3, 0, 0, 1, 1) // ADC a,x
/* 7E */ .byte opcode_clocks(3, 2, 0, 2, 1) // ROR a,x
/* 7F */ .byte opcode_clocks(4, 0, 0, 1, 1) // ADC al,x

/* 80 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BRA r
/* 81 */ .byte opcode_clocks(2, 1, 2, 1, 1) // STA (d,x)
/* 82 */ .byte opcode_clocks(3, 1, 0, 0, 1) // BRL rl
/* 83 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STA d,s
/* 84 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STY d
/* 85 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STA d
/* 86 */ .byte opcode_clocks(2, 0, 1, 0, 1) // STX d
/* 87 */ .byte opcode_clocks(2, 0, 3, 1, 1) // STA [d]

/* 88 */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEY
/* 89 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BIT i
/* 8A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXA
/* 8B */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHB
/* 8C */ .byte opcode_clocks(3, 0, 0, 1, 1) // STY a
/* 8D */ .byte opcode_clocks(3, 0, 0, 1, 1) // STA a
/* 8E */ .byte opcode_clocks(3, 0, 0, 1, 1) // STX a
/* 8F */ .byte opcode_clocks(4, 0, 0, 1, 1) // STA al

/* 90 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCC r
/* 91 */ .byte opcode_clocks(2, 1, 2, 1, 1) // STA (d),y
/* 92 */ .byte opcode_clocks(2, 0, 2, 1, 1) // STA (d)
/* 93 */ .byte opcode_clocks(2, 2, 2, 1, 1) // STA (d,s),y
/* 94 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STY d,x
/* 95 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STA d,x
/* 96 */ .byte opcode_clocks(2, 1, 1, 0, 1) // STX d,y
/* 97 */ .byte opcode_clocks(2, 0, 3, 1, 1) // STA [d],y

/* 98 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYA
/* 99 */ .byte opcode_clocks(3, 1, 0, 1, 1) // STA a,y
/* 9A */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXS
/* 9B */ .byte opcode_clocks(1, 1, 0, 0, 1) // TXY
/* 9C */ .byte opcode_clocks(3, 0, 0, 1, 1) // STZ a
/* 9D */ .byte opcode_clocks(3, 1, 0, 1, 1) // STA a,x
/* 9E */ .byte opcode_clocks(3, 1, 0, 1, 1) // STZ a,x
/* 9F */ .byte opcode_clocks(4, 0, 0, 1, 1) // STA al,x

/* A0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDY i
/* A1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // LDA (d,x)
/* A2 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDX i
/* A3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDA d,s
/* A4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDY d
/* A5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDA d
/* A6 */ .byte opcode_clocks(2, 0, 1, 0, 1) // LDX d
/* A7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // LDA [d]

/* A8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAY
/* A9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // LDA i
/* AA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TAX
/* AB */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLB
/* AC */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDY a
/* AD */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDA a
/* AE */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDX a
/* AF */ .byte opcode_clocks(4, 0, 0, 1, 1) // LDA al

/* B0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BCS r
/* B1 */ .byte opcode_clocks(2, 0, 2, 1, 1) // LDA (d),y
/* B2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // LDA (d)
/* B3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // LDA (d,s),y
/* B4 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDY d,x
/* B5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDA d,x
/* B6 */ .byte opcode_clocks(2, 1, 1, 0, 1) // LDX d,y
/* B7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // LDA [d],y

/* B8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLV
/* B9 */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDA a,y
/* BA */ .byte opcode_clocks(1, 1, 0, 0, 1) // TSX
/* BB */ .byte opcode_clocks(1, 1, 0, 0, 1) // TYX
/* BC */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDY a,x
/* BD */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDA a,x
/* BE */ .byte opcode_clocks(3, 0, 0, 1, 1) // LDX a,y
/* BF */ .byte opcode_clocks(4, 0, 0, 1, 1) // LDA al,x

/* C0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CPY i
/* C1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // CMP (d,x)
/* C2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // REP i
/* C3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // CMP d,s
/* C4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CPY d
/* C5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CMP d
/* C6 */ .byte opcode_clocks(2, 1, 2, 0, 1) // DEC d
/* C7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // CMP [d]

/* C8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INY
/* C9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CMP i
/* CA */ .byte opcode_clocks(1, 1, 0, 0, 1) // DEX
/* CB */ .byte opcode_clocks(1, 2, 0, 0, 1) // WAI
/* CC */ .byte opcode_clocks(3, 0, 0, 1, 1) // CPY a
/* CD */ .byte opcode_clocks(3, 0, 0, 1, 1) // CMP a
/* CE */ .byte opcode_clocks(3, 1, 0, 2, 1) // DEC a
/* CF */ .byte opcode_clocks(4, 0, 0, 1, 1) // CMP al

/* D0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BNE r
/* D1 */ .byte opcode_clocks(2, 0, 2, 1, 1) // CMP (d),y
/* D2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // CMP (d)
/* D3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // CMP (d,s),y
/* D4 */ .byte opcode_clocks(2, 0, 4, 0, 1) // PEI
/* D5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // CMP d,x
/* D6 */ .byte opcode_clocks(2, 2, 2, 0, 1) // DEC d,x
/* D7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // CMP [d],y

/* D8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // CLD
/* D9 */ .byte opcode_clocks(3, 0, 0, 1, 1) // CMP a,y
/* DA */ .byte opcode_clocks(1, 1, 1, 0, 1) // PHX
/* DB */ .byte opcode_clocks(1, 2, 0, 0, 1) // STP *
/* DC */ .byte opcode_clocks(3, 0, 3, 0, 1) // JML (a)
/* DD */ .byte opcode_clocks(3, 0, 0, 1, 1) // CMP a,x
/* DE */ .byte opcode_clocks(3, 2, 0, 2, 1) // DEC a,x
/* DF */ .byte opcode_clocks(4, 0, 0, 1, 1) // CMP al,x

/* E0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // CPX i
/* E1 */ .byte opcode_clocks(2, 1, 2, 1, 1) // SBC (d,x)
/* E2 */ .byte opcode_clocks(2, 1, 0, 0, 1) // SEP i
/* E3 */ .byte opcode_clocks(2, 1, 1, 0, 1) // SBC d,s
/* E4 */ .byte opcode_clocks(2, 0, 1, 0, 1) // CPX d
/* E5 */ .byte opcode_clocks(2, 0, 1, 0, 1) // SBC d
/* E6 */ .byte opcode_clocks(2, 1, 2, 0, 1) // INC d
/* E7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // SBC [d]

/* E8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // INX
/* E9 */ .byte opcode_clocks(2, 0, 0, 0, 1) // SBC i
/* EA */ .byte opcode_clocks(1, 1, 0, 0, 1) // NOP
/* EB */ .byte opcode_clocks(1, 2, 0, 0, 1) // XBA
/* EC */ .byte opcode_clocks(3, 0, 0, 1, 1) // CPX a
/* ED */ .byte opcode_clocks(3, 0, 0, 1, 1) // SBC a
/* EE */ .byte opcode_clocks(3, 1, 0, 2, 1) // INC a
/* EF */ .byte opcode_clocks(4, 0, 0, 1, 1) // SBC al

/* F0 */ .byte opcode_clocks(2, 0, 0, 0, 1) // BEQ r
/* F1 */ .byte opcode_clocks(2, 0, 2, 1, 1) // SBC (d),y
/* F2 */ .byte opcode_clocks(2, 0, 2, 1, 1) // SBC (d)
/* F3 */ .byte opcode_clocks(2, 2, 2, 1, 1) // SBC (d,s),y
/* F4 */ .byte opcode_clocks(3, 0, 2, 0, 1) // PEA
/* F5 */ .byte opcode_clocks(2, 1, 1, 0, 1) // SBC d,x
/* F6 */ .byte opcode_clocks(2, 2, 2, 0, 1) // INC d,x
/* F7 */ .byte opcode_clocks(2, 0, 3, 1, 1) // SBC [d],y

/* F8 */ .byte opcode_clocks(1, 1, 0, 0, 1) // SED
/* F9 */ .byte opcode_clocks(3, 0, 0, 1, 1) // SBC a,y
/* FA */ .byte opcode_clocks(1, 2, 1, 0, 1) // PLX
/* FB */ .byte opcode_clocks(1, 1, 0, 0, 1) // XCE
/* FC */ .byte opcode_clocks(3, 1, 2, 2, 1) // JSR (a,x) - bus access in PB
/* FD */ .byte opcode_clocks(3, 0, 0, 1, 1) // SBC a,x
/* FE */ .byte opcode_clocks(3, 2, 0, 2, 1) // INC a,x
/* FF */ .byte opcode_clocks(4, 0, 0, 1, 1) // SBC al,x

ALIGND
CPU_OpTables:
.long OpTableE0
.long OpTablePm
.long OpTablePx
.long OpTableMX

#include "CPUflags.S"

.text
ALIGNC
.globl DO_INVALID
DO_INVALID:
    decw %bx
    movl %ebx,_Map_Address
    movb %al,_Map_Byte
    jmp _InvalidOpcode     // This exits...

ALIGNC
EXPORT_C Reset_CPU
    pusha

    call _Reset_Memory__Fv

    call Invalidate_Tile_Caches

    // Reset frameskip counter
//  movl $1,FrameCount

    // Reset event timing   // 0.40
    movl $(340 * 4) - 40,_EventTrip
    movl $(340 * 4) - 40,FixedTrip
    movl $(340 * 4) - 40,_LineCycles

    // Reset bus speed
    movb $4,CycleTable

    // Reset hardware ports
    call Reset_Ports

    // Reset CGRAM
    call _Reset_CGRAM

    // Reset SPC timing
    xorl %eax,%eax
    movl %eax,SPC_last_cycles
    movl %eax,SPC_CPU_cycles
    movl %eax,SPC_cycles_left

    // Reset CPU
    movb %al,CPU_Executing
    movb %al,In_WAI
    movl $OpTableE1,OpTable // Set current opcode emulation table
    movl %eax,SNES_Cycles   // Clear cycle counts
    movl %eax,_Cycles       // Needed for IRQNewFrame later
    movl %eax,_A            // Clear A, D, X, Y
    movl %eax,_D
    movl %eax,_X
    movl %eax,_Y
    movl %eax,_DB_Shifted   // Clear bank registers
    movl %eax,_PB_Shifted
    movb %al,_OLD_PB

#ifdef DEBUG
    movl %eax,_Frames
    movl %eax,_Timer_Counter_FPS
#endif
    // Initialize flags
    FLAGS_TO (SNES_FLAG_B1+SNES_FLAG_E+SNES_FLAG_I)
    movl _Read_BankMapping,%eax
    movl %eax,_PB_Read  // Bank 0 read through memory mapper
    movl $0x01FF,_S

    // Get all interrupt vectors
    movl _RomAddress,%ebx       // Get address of ROM
    movzwl 0xFFFC(%ebx),%eax    // Get Reset vector
    xorl %edx,%edx
    movl %eax,_RES_Evector      // Save Reset vector
    movl %eax,_PC               // Setup PC
    movl %eax,_OLD_PC
    addb %ah,%ah
    adcl %edx,%edx
    movl _Read_BankOffset(,%edx,4),%eax
    movl %eax,_RES_Eoffset
    movl %eax,_PBOffset

    movzwl 0xFFEA(%ebx),%eax    // Get Native mode NMI vector
    xorl %edx,%edx
    movl %eax,_NMI_Nvector
    addb %ah,%ah
    adcl %edx,%edx
    movl _Read_BankOffset(,%edx,4),%eax
    movl %eax,_NMI_Noffset

    movzwl 0xFFFA(%ebx),%eax    // Get Emulation mode NMI vector
    xorl %edx,%edx
    movl %eax,_NMI_Evector
    addb %ah,%ah
    adcl %edx,%edx
    movl _Read_BankOffset(,%edx,4),%eax
    movl %eax,_NMI_Eoffset

    movzwl 0xFFEE(%ebx),%eax    // Get Native mode IRQ vector
    xorl %edx,%edx
    movl %eax,_IRQ_Nvector
    addb %ah,%ah
    adcl %edx,%edx
    movl _Read_BankOffset(,%edx,4),%eax
    movl %eax,_IRQ_Noffset

    movzwl 0xFFFE(%ebx),%eax    // Get Emulation mode BRK/IRQ vector
    xorl %edx,%edx
    movl %eax,_IRQ_Evector
    addb %ah,%ah
    adcl %edx,%edx
    movl _Read_BankOffset(,%edx,4),%eax
    movl %eax,_IRQ_Eoffset

    movzwl 0xFFE6(%ebx),%eax    // Get Native mode BRK vector
    xorl %edx,%edx
    movl %eax,_BRK_Nvector
    addb %ah,%ah
    adcl %edx,%edx
    movl _Read_BankOffset(,%edx,4),%eax
    movl %eax,_BRK_Noffset

    movzwl 0xFFE4(%ebx),%eax    // Get Native mode COP vector
    xorl %edx,%edx
    movl %eax,_COP_Nvector
    addb %ah,%ah
    adcl %edx,%edx
    movl _Read_BankOffset(,%edx,4),%eax
    movl %eax,_COP_Noffset

    movzwl 0xFFF4(%ebx),%eax    // Get Emulation mode COP vector
    xorl %edx,%edx
    movl %eax,_COP_Evector
    addb %ah,%ah
    adcl %edx,%edx
    movl _Read_BankOffset(,%edx,4),%eax
    movl %eax,_COP_Eoffset

    // Reset renderer
    movb $0xFF,TM_MASK

/*  movb _SPC_ENABLED,%al
    pushl %eax
    movb $0,_SPC_ENABLED    // If SPC is enabled, it will execute!
    call IRQNewFrame        // This relies on other CPU bits to be reset!
    popl %eax
    movb %al,_SPC_ENABLED*/
    call IRQNewFrameReset   // This relies on other CPU bits to be reset!

    popa
    ret

ALIGNC
EXPORT_C Do_CPU
    pusha
    movb $1,PaletteChanged  // Make sure we get our palette
    call CPU_START
1:  popa
    ret

/* Start of actual CPU execution core */

// New for 0.25 - one CPU execution loop, also used for SPC
ALIGNC
.globl CPU_START_IRQ
CPU_START_IRQ:
    CheckVIRQ   // Check for Vertical IRQ
CPU_START:
#ifdef DEBUG
    movl _Timer_Counter_Profile,%eax
    movl %eax,_CPU_Profile_Last
#endif
    LOAD_P
    LOAD_PC
    LOAD_CYCLES
    LOAD_BASE
    xorl %eax,%eax          // Zero for table offset
    movb $-1,CPU_Executing
    jmp CPU_START_NEXT

ALIGNC
EXPORT_C CPU_RETURN
CPU_RETURN:
#ifndef POST_ADD_CYCLES
    popl %ebx
    movl R_EventTrip,%edi
    addl %ebx,_Cycles   // Update cycle counter
#else
    movl R_EventTrip,%edi
#endif
    xorl %eax,%eax          // Zero for table offset

    cmpl %edi,_Cycles
    jnb HANDLE_EVENT
EXPORT_C CPU_START_NEXT
CPU_START_NEXT:
// This code is for a CPU-tracker dump... #define TRACKERS to make a dump
// of the CPU state before each instruction - uncomment the ret to
// force emulation core to break when buffer fills. TRACKERS must be
// defined to the size of the buffer to be used - which must be a power
// of two, and the variables required by this and the write in Wangle()
// (main.cc) exist only if DEBUG and TRACKERS are also defined in main.cc
// and romload.cc.
#ifdef TRACKERS
#if TRACKERS >= 16
    movl _LastIns,%edi      //
    addl _InsAddress,%edi   //
    movb _PB,%al            //
    movb %al,(%edi)         //
    GET_PC %eax             //
    movb %ah,1(%edi)        //
    movb %al,2(%edi)        //
    movb _A+1,%al           //
    movb %al,3(%edi)        //
    movb _A,%al             //
    movb %al,4(%edi)        //
    movb _X+1,%al           //
    movb %al,5(%edi)        //
    movb _X,%al             //
    movb %al,6(%edi)        //
    movb _Y+1,%al           //
    movb %al,7(%edi)        //
    movb _Y,%al             //
    movb %al,8(%edi)        //
    movb _S+1,%al           //
    movb %al,9(%edi)        //
    movb _S,%al             //
    movb %al,10(%edi)       //
    movb _DB,%al            //
    movb %al,11(%edi)       //
    movb _D+1,%al           //
    movb %al,12(%edi)       //
    movb _D,%al             //
    movb %al,13(%edi)       //
    SAVE_P                  //
    movb _P+1,%al           //
    andb $1,%al             //
    movb %al,15(%edi)       //
    jnz 0f                  //
    E0_SETUPFLAGS           //
    jmp 1f                  //
0:  E1_SETUPFLAGS           //
1:  movb %al,14(%edi)       //
    movl _LastIns,%edi      //
    addl $16,%edi           //
    andl $(TRACKERS-1),%edi //
    movl %edi,_LastIns      //
    testl %edi,%edi         //
    jnz 0f                  //
//  ret                     //
0:                          //
#endif
#endif
    movl R_OpTable,%edi
    movb (_NativePC),%al    // Get opcode
    movl %eax,%ebx

    // Bus speed selected by modifying code
    movb 0x400(%edi,%eax),%bl   // SlowROM, timing * 8, FastROM, timing * 6
.globl CycleTable
CycleTable=.-3  // Address of '04'

#ifndef POST_ADD_CYCLES
    pushl %ebx
#else
    addl %ebx,_Cycles   // Update cycle counter
#endif

    jmpl (%edi,%eax,4)      // Call opcode handler

ALIGNC
HANDLE_EVENT:
    SAVE_P
    SAVE_PC %ebx
  
    movb $0,CPU_Executing

#ifdef DEBUG
    movl _Timer_Counter_Profile,%eax
    subl _CPU_Profile_Last,%eax
    addl %eax,_CPU_Profile_Test
#endif

    SAVE_CYCLES
    jmpl *Event_Handler

#include "CPUaddr.S"    // Addressing mode macros ([d], a,x, etc.)
#include "CPUmacro.S"   // Instruction macros (LDA,ADC,SBC,etc.)

EXPORT_C cpu_ops_start

#include "CPU0.S"       // Opcode handlers for: 0x00-0x0F
#include "CPU1.S"       // Opcode handlers for: 0x10-0x1F
#include "CPU2.S"       // Opcode handlers for: 0x20-0x2F
#include "CPU3.S"       // Opcode handlers for: 0x30-0x3F
#include "CPU4.S"       // Opcode handlers for: 0x40-0x4F
#include "CPU5.S"       // Opcode handlers for: 0x50-0x5F
#include "CPU6.S"       // Opcode handlers for: 0x60-0x6F
#include "CPU7.S"       // Opcode handlers for: 0x70-0x7F
#include "CPU8.S"       // Opcode handlers for: 0x80-0x8F
#include "CPU9.S"       // Opcode handlers for: 0x90-0x9F
#include "CPUA.S"       // Opcode handlers for: 0xA0-0xAF
#include "CPUB.S"       // Opcode handlers for: 0xB0-0xBF
#include "CPUC.S"       // Opcode handlers for: 0xC0-0xCF
#include "CPUD.S"       // Opcode handlers for: 0xD0-0xDF
#include "CPUE.S"       // Opcode handlers for: 0xE0-0xEF
#include "CPUF.S"       // Opcode handlers for: 0xF0-0xFF

#include "timing.S"

.text; ALIGNC
.data; ALIGND
.bss; ALIGND
